"""Submodule for NeuroKit."""

from .ecg_simulate import ecg_simulate
from .ecg_clean import ecg_clean
from .ecg_findpeaks import ecg_findpeaks
from .ecg_rate import ecg_rate
from .ecg_process import ecg_process
from .ecg_plot import ecg_plot



# -*- coding: utf-8 -*-
import numpy as np
import pandas as pd

from ..signal import signal_interpolate


def ecg_rate(peaks, sampling_rate=1000, desired_length=None):
    """Calculate heart rate from R-peaks.

    Parameters
    ----------
    peaks : list, array, DataFrame, Series or dict
        The samples at which the R-peaks occur. If a dict or a DataFrame is
        passed, it is assumed that these containers were obtained with
        `ecg_findpeaks()`.
    sampling_rate : int
        The sampling frequency of the signal that contains the R-peaks (in Hz,
        i.e., samples/second). Defaults to 1000.
    desired_length : int
        By default, the returned heart rate has the same number of elements as
        peaks. If set to an integer, the returned heart rate will be
        interpolated between R-peaks over `desired_length` samples. Has no
        effect if a DataFrame is passed in as the `peaks` argument. Defaults to
        None.

    Returns
    -------
    signals : DataFrame
        A DataFrame containing heart rate accessible with the key "ECG_Rate".

    See Also
    --------
    ecg_clean, ecg_findpeaks, ecg_process, ecg_plot

    Examples
    --------
    >>> import neurokit2 as nk
    >>>
    >>> ecg = nk.ecg_simulate(duration=15, heart_rate=80)
    >>> signals, info = nk.ecg_findpeaks(ecg)
    >>>
    >>> data = nk.ecg_rate(signals)
    >>> data["ECG_Signal"] = ecg  # Add the signal back
    >>> data.plot(subplots=True)
    """
    if isinstance(peaks, dict):
        peaks = peaks["ECG_Peaks"]
    elif isinstance(peaks, pd.DataFrame):
        desired_length = len(peaks["ECG_Peaks"])
        peaks = np.where(peaks["ECG_Peaks"] == 1)[0]

    # Determine length of final signal to return.
    if desired_length is None:
        desired_length = len(peaks)

    # Sanity checks.
    if len(peaks) <= 3:
        print("NeuroKit warning: ecg_rate(): too few peaks detected to "
              "compute the rate. Returning empty variable(s).")
        return pd.DataFrame({"ECG_Rate": np.full(desired_length, np.nan)})


    # Calculate period in msec, based on peak to peak difference and make sure
    # that rate has the same number of elements as peaks (important for
    # interpolation later) by prepending the mean of all periods.
    period = np.ediff1d(peaks, to_begin=0) / sampling_rate
    period[0] = np.mean(period)
    # Get rate.
    rate = 60 / period

    # Interpolate all statistics to desired lenght.
    rate = signal_interpolate(rate, x_axis=peaks,
                              desired_length=desired_length)

    # Prepare output
    signals = pd.DataFrame(rate, columns=["ECG_Rate"])
    return signals



# -*- coding: utf-8 -*-
import numpy as np
import pandas as pd

import scipy.sparse

from ..stats import loess


def signal_detrend(signal, method="polynomial", order=1, regularization=500, alpha=0.75):
    """Polynomial detrending of signal.

    Apply a baseline (order = 0), linear (order = 1), or polynomial (order > 1) detrending to the signal (i.e., removing a general trend).

    Parameters
    ----------
    signal : list, array or Series
        The signal channel in the form of a vector of values.
    method : str
        Can be one of 'polynomial' (default; traditional detrending of a given order) or 'tarvainen2002' to use the smoothness priors approach described by Tarvainen (2002) (mostly used in HRV analyses as a lowpass filter to remove complex trends), or 'loess' for LOESS smoothing trend removal.
    order : int
        The order of the polynomial. 0, 1 or > 1 for a baseline ('constant detrend', i.e., remove only the mean), linear (remove the linear trend) or polynomial detrending.
    regularization : int
        Only used if `method='tarvainen2002'`. The regularization parameter (default to 500).
    alpha : float
        Only used if `method` is 'loess'. The parameter which controls the degree of smoothing.

    Returns
    -------
    array
        Vector containing the detrended signal.

    See Also
    --------
    signal_filter, loess

    Examples
    --------
    >>> import numpy as np
    >>> import pandas as pd
    >>> import neurokit2 as nk
    >>> import matplotlib.pyplot as plt
    >>>
    >>> signal = np.cos(np.linspace(start=0, stop=10, num=1000))  # Low freq
    >>> signal += np.cos(np.linspace(start=0, stop=100, num=1000))  # High freq
    >>> signal += 3  # Add baseline
    >>>
    >>> axes = pd.DataFrame({"Raw": signal,
                      "Baseline": nk.signal_detrend(signal, order=0),
                      "Linear": nk.signal_detrend(signal, order=1),
                      "Quadratic": nk.signal_detrend(signal, order=2),
                      "Cubic": nk.signal_detrend(signal, order=3),
                      "10th": nk.signal_detrend(signal, order=10),
                      "Tarvainen": nk.signal_detrend(signal, method='tarvainen2002'),
                      "LOESS": nk.signal_detrend(signal, method='loess')}).plot(subplots=True)
    >>> # Plot horizontal lines to better visualize the detrending
    >>> for subplot in axes:
    >>>     subplot.axhline(y=0, color='k', linestyle='--')

    References
    ----------
    - `Tarvainen, M. P., Ranta-Aho, P. O., & Karjalainen, P. A. (2002). An advanced detrending method with application to HRV analysis. IEEE Transactions on Biomedical Engineering, 49(2), 172-175. <https://ieeexplore.ieee.org/document/979357>`_
    """
    method = method.lower()
    if method in ["tarvainen", "tarvainen2002"]:
        detrended = _signal_detrend_tarvainen2002(signal, regularization)
    elif method in ["poly", "polynomial"]:
        detrended = _signal_detrend_polynomial(signal, order)
    elif method in ["loess", "lowess"]:
        detrended = _signal_detrend_loess(signal, alpha=alpha)
    else:
        raise ValueError("NeuroKit error: signal_detrend(): 'method' should be "
                         "one of 'polynomial', 'loess' or 'tarvainen2002'.")

    return detrended




# =============================================================================
# Internals
# =============================================================================
def _signal_detrend_loess(signal, alpha=0.75):
    detrended = np.array(signal) - loess(signal, alpha=alpha)
    return detrended





def _signal_detrend_polynomial(signal, order=1):
    x_axis = np.linspace(0, 100, num=len(signal))

    # Generating weights and model for polynomial function with a given degree
    trend = np.polyval(np.polyfit(x_axis, signal, order), x_axis)

    # detrend
    detrended = np.array(signal) - trend
    return detrended




def _signal_detrend_tarvainen2002(signal, regularization=500):
    """`Tarvainen, M. P., Ranta-Aho, P. O., & Karjalainen, P. A. (2002). An advanced detrending method with application to HRV analysis. IEEE Transactions on Biomedical Engineering, 49(2), 172-175. <https://ieeexplore.ieee.org/document/979357>`_
    """
    N = len(signal)
    identity = np.eye(N)
    B = np.dot(np.ones((N-2, 1)), np.array([[1, -2, 1]]))
    D_2 = scipy.sparse.dia_matrix((B.T, [0, 1, 2]), shape=(N-2, N))
    inv = np.linalg.inv(identity + regularization**2 * D_2.T @ D_2)
    z_stat = ((identity - inv)) @ signal

    trend = np.squeeze(np.asarray(signal - z_stat))

    # detrend
    detrended = np.array(signal) - trend
    return detrended



"""Submodule for NeuroKit."""

from .entropy_shannon import entropy_shannon
from .entropy_approximate import entropy_approximate
from .entropy_sample import entropy_sample
from .entropy_fuzzy import entropy_fuzzy



# -*- coding: utf-8 -*-
import numpy as np
import pandas as pd

import scipy.signal
import scipy.misc

from ..stats import standardize




def signal_findpeaks(signal, distance_min=None, height_min=None, width_min=None, distance_max=None, height_max=None, width_max=None, relative_distance_min=None, relative_height_min=None, relative_width_min=None, relative_distance_max=None, relative_height_max=None, relative_width_max=None):
    """Find peaks in a signal.

    Locate peaks (local maxima) in a signal and their related characteristics, such as height (prominence), width and distance with other peaks.

    Parameters
    ----------
    signal : list, array or Series
        The signal channel in the form of a vector of values.
    distance_min, height_min, width_min, distance_max, height_max, width_max : float
        The minimum or maximum distance (between peaks, in number of sample points), height (i.e., amplitude in terms of absolute values) or width of the peaks (in number of sample points). For example, `distance_min=20` will remove all peaks which distance with the previous peak is smaller or equal to 20 sample points.
    relative_distance_min, relative_height_min, relative_width_min, relative_distance_max, relative_height_max, relative_width_max : float
        The minimum or maximum distance (between peaks), height (i.e., amplitude) or width of the peaks in terms of standard deviation from the sample. For example, `relative_distance_min=-2.96` will remove all peaks which distance lies below 2.96 standard deviations from the mean of the distances.

    Returns
    ----------
    array, dict
        Returns two things. An array containing the peaks indices (as relative to the given signal).
        For instance, the value 3 means that the third sample of the signal is a peak or a troughs.
        It also returns a dict itself containing 3 arrays:
            - 'Distance' contains, for each peak, the closest distance with another peak. Note that these values will be recomputed after filtering to match the selected peaks.
            - 'Height' contains the prominence of each peak. See `scipy.signal.peak_prominences()`.
            - 'Width' contains the width of each peak. See `scipy.signal.peak_widths()`.

    Examples
    ---------
    >>> import numpy as np
    >>> import pandas as pd
    >>> import neurokit2 as nk
    >>> import scipy.misc
    >>>
    >>> signal = np.cos(np.linspace(start=0, stop=30, num=1000))
    >>> pd.Series(signal).plot()
    >>> peaks, info = nk.signal_findpeaks(signal)
    >>> nk.plot_events_in_signal(signal, peaks)
    >>>
    >>> signal = np.concatenate([np.arange(0, 20, 0.1), np.arange(17, 30, 0.1), np.arange(30, 10, -0.1)])
    >>> peaks, info = nk.signal_findpeaks(signal)
    >>> nk.plot_events_in_signal(signal, peaks)
    >>>
    >>> # Filter peaks
    >>> ecg = scipy.misc.electrocardiogram()
    >>> signal = ecg[0:1000]
    >>> peaks, info = nk.signal_findpeaks(signal, relative_height_min=0)
    >>> peaks2, info = nk.signal_findpeaks(signal, relative_height_min=1)
    >>> nk.plot_events_in_signal(signal, [peaks, peaks2])

    See Also
    --------
    scipy.signal.find_peaks, scipy.signal.peak_widths, peak_prominences.signal.peak_widths
    """
    peaks, info = _signal_findpeaks(signal)

    keep = np.full(len(peaks), True)

    # Absolute indices - min
    if distance_min is not None:
        keep[info["Distance"] < distance_min] = False
    if height_min is not None:
        keep[info["Height"] < height_min] = False
    if width_min is not None:
        keep[info["Width"] < width_min] = False

    # Absolute indices - max
    if distance_max is not None:
        keep[info["Distance"] > distance_max] = False
    if height_max is not None:
        keep[info["Height"] > height_max] = False
    if width_max is not None:
        keep[info["Width"] > width_max] = False

    # Relative indices - min
    if relative_distance_min is not None:
        keep[standardize(info["Distance"]) < relative_distance_min] = False
    if relative_height_min is not None:
        keep[standardize(info["Height"]) < relative_height_min] = False
    if relative_width_min is not None:
        keep[standardize(info["Width"]) < relative_width_min] = False

    # Relative indices - max
    if relative_distance_max is not None:
        keep[standardize(info["Distance"]) > relative_distance_max] = False
    if relative_height_max is not None:
        keep[standardize(info["Height"]) > relative_height_max] = False
    if relative_width_max is not None:
        keep[standardize(info["Width"]) > relative_width_max] = False

    # Filter
    peaks = peaks[keep]
    info["Distance"] = _signal_findpeaks_distances(peaks)
    info["Height"] = info["Height"][keep]
    info["Width"] = info["Width"][keep]

    return peaks, info





# =============================================================================
# Internals
# =============================================================================


def _signal_findpeaks_distances(peaks):

    if len(peaks) <= 2:
        distances = np.full(len(peaks), np.nan)
    else:
        distances_next = np.concatenate([[np.nan], np.abs(np.diff(peaks))])
        distances_prev = np.concatenate([np.abs(np.diff(peaks[::-1])), [np.nan]])
        distances = np.array([np.nanmin(i) for i in list(zip(distances_next, distances_prev))])

    return distances






def _signal_findpeaks(signal):
    peaks, _ = scipy.signal.find_peaks(signal)

    # Get info
    distances = _signal_findpeaks_distances(peaks)
    heights, left_base, right_base = scipy.signal.peak_prominences(signal, peaks)
    widths, width_heights, left_ips, right_ips = scipy.signal.peak_widths(signal, peaks, rel_height=0.5)

    # Prepare output
    info = {"Distance": distances,
            "Height": heights,
            "Width": widths}

    return peaks, info



# -*- coding: utf-8 -*-
import numpy as np


def _embed(signal, order=3, delay=1):
    """Time-delay embedding.

    Adapted from `EntroPy <https://github.com/raphaelvallat/entropy>`_, check it out!

    Parameters
    ----------
    signal : list, array or Series
        The signal channel in the form of a vector of values.
    order : int
        Embedding dimension (order).
    delay : int
        Delay.

    Returns
    -------
    array
        Embedded time-series, of shape (n_times - (order - 1) * delay, order)

    Examples
    ---------
    >>> signal = np.cos(np.linspace(start=0, stop=20, num=1000))
    >>> _embed(signal)
    """
    N = len(signal)

    # Sanity checks
    if order * delay > N:
        raise ValueError("NeuroKit error: _embedding(): order * delay should be lower than length of signal.")
    if delay < 1:
        raise ValueError("NeuroKit error: _embedding(): 'delay' has to be at least 1.")
    if order < 2:
        raise ValueError("NeuroKit error: _embedding(): 'order' has to be at least 2.")

    Y = np.zeros((order, N - (order - 1) * delay))
    for i in range(order):
        Y[i] = signal[i * delay:i * delay + Y.shape[1]]

    return(Y.T)



# -*- coding: utf-8 -*-
import numpy as np

def signal_zerocrossings(signal):
    """Locate the indices where the signal crosses zero.

    Note that when the signal crosses zero between two points, the first index is returned.

    Parameters
    ----------
    signal : list, array or Series
        The signal channel in the form of a vector of values.

    Returns
    -------
    array
        Vector containing the indices of zero crossings.

    Examples
    --------
    >>> import numpy as np
    >>> import pandas as pd
    >>> import neurokit2 as nk
    >>>
    >>> signal = np.cos(np.linspace(start=0, stop=15, num=1000))
    >>> zeros = nk.signal_zerocrossings(signal)
    >>> nk.events_plot(zeros, signal)
    """
    df = np.diff(np.sign(signal))
    zeros = np.nonzero(np.abs(df) > 0)[0]

    return zeros



# -*- coding: utf-8 -*-
import numpy as np
import pandas as pd

import scipy.linalg






def loess(y, X=None, alpha=0.75, order=2):
    """Local Polynomial Regression (LOESS)

    Performs a LOWESS (LOcally WEighted Scatter-plot Smoother) regression.


    Parameters
    ----------
    y : list, array or Series
        The response variable (the y axis).
    X : list, array or Series
        Explanatory variable (the x axis). If 'None', will treat y as a continuous signal (useful for smoothing).
    alpha : float
        The parameter which controls the degree of smoothing, which corresponds
        to the proportion of the samples to include in local regression.
    order : int
        Degree of the polynomial to fit. Can be 1 or 2 (default).

    Returns
    -------
    array
        Prediciton of the LOESS algorithm.

    See Also
    ----------
    signal_smooth, signal_detrend

    Examples
    ---------
    >>> import pandas as pd
    >>> import neurokit2 as nk
    >>>
    >>> signal = np.cos(np.linspace(start=0, stop=10, num=1000))
    >>> distorted = nk.signal_distord(signal, noise_amplitude=[0.3, 0.2, 0.1], noise_frequency=[5, 10, 50])
    >>>
    >>> pd.DataFrame({
            "Raw": distorted,
            "Loess_1": nk.loess(distorted, order=1),
            "Loess_2": nk.loess(distorted, order=2)}).plot()

    References
    ----------
    - https://simplyor.netlify.com/loess-from-scratch-in-python-animation.en-us/
    """
    if X is None:
        X = np.linspace(0, 1, len(y))


    assert (order == 1) or (order == 2), "Deg has to be 1 or 2"
    assert (alpha > 0) and (alpha <= 1), "Alpha has to be between 0 and 1"
    assert len(X) == len(y), "Length of X and y are different"


    X_domain = X

    n = len(X)
    span = int(np.ceil(alpha * n))

    y_hat = np.zeros(len(X_domain))
    x_space = np.zeros_like(X_domain)

    for i, val in enumerate(X_domain):
        distance = abs(X - val)
        sorted_dist = np.sort(distance)
        ind = np.argsort(distance)

        Nx = X[ind[:span]]
        Ny = y[ind[:span]]

        delx0 = sorted_dist[span-1]

        u = distance[ind[:span]] / delx0
        w = (1 - u**3)**3

        W = np.diag(w)
        A = np.vander(Nx, N=1+order)

        V = np.matmul(np.matmul(A.T, W), A)
        Y = np.matmul(np.matmul(A.T, W), Ny)
        Q, R = scipy.linalg.qr(V)
        p = scipy.linalg.solve_triangular(R, np.matmul(Q.T, Y))

        y_hat[i] = np.polyval(p, val)
        x_space[i] = val

    return y_hat



# -*- coding: utf-8 -*-
import pandas as pd
import numpy as np
import scipy.stats

from ..signal import signal_resample


def ppg_simulate(duration=10, length=None, sampling_rate=1000, noise=0.01, heart_rate=60):
    """Simulate a PPG (photoplethysmography) signal

    Generate an artificial (synthetic) PPG signal of a given duration and sampling rate.

    Parameters
    ----------
    duration : int
        Desired recording length in seconds.
    sampling_rate, length : int
        The desired sampling rate (in Hz, i.e., samples/second) or the desired
        length of the signal (in samples).
    noise : float
       Noise level (gaussian noise).
    heart_rate : int
        Desired simulated heart rate (in beats per minute).

    Returns
    ----------
   array
        Array containing the PPG signal.

    Example
    ----------
    >>> import neurokit as nk
    >>> import pandas as pd
    >>>
    >>> ppg = nk.ppg_simulate(duration=10, n_bursts=3)
    >>> pd.Series(ppg).plot()

    See Also
    --------
    signal_resample, ecg_simulate, emg_simulate


    References
    ----------
    - `Banerjee et al. (2015). Noise cleaning and Gaussian modeling of smart phone photoplethysmogram to improve blood pressure estimation. In 2015 IEEE International Conference on Acoustics, Speech and Signal Processing (ICASSP) (pp. 967-971). IEEE. <http://150.162.46.34:8080/icassp2015/pdfs/0000967.pdf>`_.
    """
    index = np.linspace(-9, 9, num=100)
    normal1 = scipy.stats.norm.pdf(index, loc=-2, scale=2)/1.5
    normal2 = scipy.stats.norm.pdf(index, loc=2, scale=1.5)
    ppg = normal1+normal2
    pd.Series(ppg).plot()
    return(ppg)







# -*- coding: utf-8 -*-
import pandas as pd
import numpy as np

from .signal_resample import signal_resample


def signal_merge(signal1, signal2, time1=[0, 10], time2=[0, 10]):
    """Arbitrary addition of two signals with different time ranges.

    Parameters
    ----------
    signal1, signal2 : list, array or Series
        The signal channels in the form of a vector of values.
    time1, time2 : list
        Lists containing two numeric values corresponding to the beginning and end of 'signal1' and 'signal2', respectively.

    Returns
    -------
    array
        Vector containing the sum of the two signals.

    Examples
    --------
    >>> import numpy as np
    >>> import pandas as pd
    >>> import neurokit2 as nk
    >>>
    >>> signal1 = np.cos(np.linspace(start=0, stop=10, num=100))
    >>> signal2 = np.cos(np.linspace(start=0, stop=20, num=100))
    >>>
    >>> signal = nk.signal_merge(signal1, signal2, time1=[0, 10], time2=[-5, 5])
    >>> nk.signal_plot(signal)
    """

    # Resample signals if different
    sampling_rate1 = len(signal1) / np.diff(time1)[0]
    sampling_rate2 = len(signal2) / np.diff(time2)[0]
    if sampling_rate1 > sampling_rate2:
        signal2 = signal_resample(signal2, sampling_rate=sampling_rate2, desired_sampling_rate=sampling_rate1)
    elif sampling_rate2 > sampling_rate1:
        signal1 = signal_resample(signal1, sampling_rate=sampling_rate1, desired_sampling_rate=sampling_rate2)
    sampling_rate = np.max([sampling_rate1, sampling_rate2])

    # Fill beginning
    if time1[0] < time2[0]:
        beginning = np.full(int(np.round(sampling_rate*(time2[0] - time1[0]))), signal2[0])
        signal2 = np.concatenate((beginning, signal2))
    elif time2[0] < time1[0]:
        beginning = np.full(int(np.round(sampling_rate*(time1[0] - time2[0]))), signal1[0])
        signal1 = np.concatenate((beginning, signal1))

    # Fill end
    if time1[1] > time2[1]:
        end = np.full(int(np.round(sampling_rate*(time1[1] - time2[1]))), signal2[-1])
        signal2 = np.concatenate((signal2, end))
    elif time2[1] > time1[1]:
        end = np.full(int(np.round(sampling_rate*(time2[1] - time1[1]))), signal1[-1])
        signal1 = np.concatenate((signal1, end))

    # Sanitize length of arrays
    if len(signal1) > len(signal2):
        signal1 = signal1[0:len(signal2)]
    if len(signal2) > len(signal1):
        signal2 = signal2[0:len(signal1)]

    merged = signal1 + signal2
    return(merged)



# -*- coding: utf-8 -*-
import pandas as pd
import numpy as np





def _signal_binarize(signal, threshold="auto"):
    if threshold == "auto":
        threshold = np.mean([np.max(signal), np.min(signal)])

    binary = signal.copy()
    binary[signal > threshold] = 1
    binary[signal <= threshold] = 0
    return(binary)







def signal_binarize(signal, threshold="auto"):
    """Binarize a continuous signal.

    Convert a continuous signal into zeros and ones depending on a given threshold.

    Parameters
    ----------
    signal : list, array or Series
        The signal channel in the form of a vector of values.
    threshold : float
        The threshold value by which to select the events. If "auto", takes the value between the max and the min.

    Returns
    -------
    list
        A list or array depending on the type passed.

    Examples
    --------
    >>> import numpy as np
    >>> import pandas as pd
    >>> import neurokit2 as nk
    >>>
    >>> signal = np.cos(np.linspace(start=0, stop=20, num=1000))
    >>> binary = nk.signal_binarize(signal)
    >>> pd.DataFrame({"Raw": signal, "Binary": binary}).plot()
    """

    # Return appropriate type
    if isinstance(signal, list):
        binary = _signal_binarize(np.array(signal), threshold=threshold)
        signal = list(binary)
    elif isinstance(signal, pd.Series):
        binary = _signal_binarize(signal.values, threshold=threshold)
        signal[:] = binary
    else:
        signal = _signal_binarize(signal, threshold=threshold)

    return(signal)





# -*- coding: utf-8 -*-
from ..signal import signal_detrend
from ..signal import signal_filter


def rsp_clean(rsp_signal, sampling_rate=1000, method="khodadad2018"):
    """Preprocess a respiration (RSP) signal.

    Clean a respiration signal using different sets of parameters, such as:

    - `Khodadad et al. (2018)
    <https://iopscience.iop.org/article/10.1088/1361-6579/aad7e6/meta>`_:
        linear detrending followed by a fifth order 2Hz low-pass IIR
        Butterworth filter).
    - `BioSPPy
    <https://github.com/PIA-Group/BioSPPy/blob/master/biosppy/signals/resp.py>`_:
        second order 0.1 - 0.35 Hz bandpass Butterworth filter followed by a
        constant detrending.

    Parameters
    ----------
    rsp_signal : list, array or Series
        The raw respiration channel (as measured, for instance, by a
        respiration belt).
    sampling_rate : int
        The sampling frequency of `rsp_signal` (in Hz, i.e., samples/second).
    method : str
        The processing pipeline to apply. Can be one of "khodadad2018"
        (default) or "biosppy".

    Returns
    -------
    array
        Vector containing the cleaned respiratory signal.

    See Also
    --------
    rsp_findpeaks, rsp_rate, rsp_amplitude, rsp_process, rsp_plot

    Examples
    --------
    >>> import pandas as pd
    >>> import neurokit2 as nk
    >>>
    >>> rsp = nk.rsp_simulate(duration=30, sampling_rate=50, noise=0.01)
    >>> signals = pd.DataFrame({
            "RSP_Raw": rsp,
            "RSP_Khodadad2018": nk.rsp_clean(rsp, sampling_rate=50, method="khodadad2018"),
            "RSP_BioSPPy": nk.rsp_clean(rsp, sampling_rate=50, method="biosppy")})
    >>> signals.plot()
    """
    method = method.lower()  # remove capitalised letters
    if method in ["khodadad", "khodadad2018"]:
        clean = _rsp_clean_khodadad2018(rsp_signal, sampling_rate)
    elif method == "biosppy":
        clean = _rsp_clean_biosppy(rsp_signal, sampling_rate)
    else:
        raise ValueError("NeuroKit error: rsp_clean(): 'method' should be "
                         "one of 'khodadad2018' or 'biosppy'.")

    return clean


# =============================================================================
# Khodadad et al. (2018)
# =============================================================================
def _rsp_clean_khodadad2018(rsp_signal, sampling_rate=1000):
    """The algorithm is based on (but not an exact
    implementation of) the "Zero-crossing algorithm with amplitude threshold"
    by `Khodadad et al. (2018)
    <https://iopscience.iop.org/article/10.1088/1361-6579/aad7e6/meta>`_.
    """
    # Detrend and lowpass-filter the signal to be able to reliably detect
    # zero crossings in raw signal.
    clean = signal_detrend(rsp_signal, order=1)
    clean = signal_filter(clean, sampling_rate=sampling_rate,
                          lowcut=None, highcut=2,
                          method="butterworth", order=5)

    return clean


# =============================================================================
# BioSPPy
# =============================================================================
def _rsp_clean_biosppy(rsp_signal, sampling_rate=1000):
    """Uses the same defaults as `BioSPPy
    <https://github.com/PIA-Group/BioSPPy/blob/master/biosppy/signals/resp.py>`_.
    """
    clean = signal_filter(rsp_signal, sampling_rate=sampling_rate,
                          lowcut=0.1, highcut=0.35,
                          method="butterworth", order=2)
    clean = signal_detrend(clean, order=0)

    return clean



# -*- coding: utf-8 -*-
import pandas as pd
import numpy as np

import scipy.signal
import scipy.ndimage



def signal_resample(signal, desired_length=None, sampling_rate=None, desired_sampling_rate=None, method="interpolation"):
    """Resample a continuous signal to a different length or sampling rate.

    Up- or down-sample a signal. The user can specify either a desired length for the vector, or input the original sampling rate and the desired sampling rate. See https://github.com/neuropsychology/NeuroKit/scripts/resampling.ipynb for a comparison of the methods.

    Parameters
    ----------
    signal :  list, array or Series
        The signal channel in the form of a vector of values.
    desired_length : int
        The desired length of the signal.
    sampling_rate, desired_sampling_rate : int
        The original and desired (output) sampling frequency (in Hz, i.e., samples/second).
    method : str
        Can be 'numpy' (default) for numpy's interpolation (see `numpy.interp()`), 'pandas' for Pandas' time series resampling, 'interpolation' (see `scipy.ndimage.zoom()`), 'poly' (see `scipy.signal.resample_poly()`) or 'FFT' (see `scipy.signal.resample()`) for the Fourier method. FFT is the most accurate (if the signal is periodic), but becomes exponentially slower as the signal length increases. In contrast, 'numpy' is the fastest, followed by 'poly', 'pandas' and 'interpolation'.

    Returns
    -------
    array
        Vector containing resampled signal values.

    Examples
    --------
    >>> import numpy as np
    >>> import pandas as pd
    >>> import neurokit2 as nk
    >>>
    >>> signal = np.cos(np.linspace(start=0, stop=20, num=100))
    >>>
    >>> # Downsample
    >>>
    >>> downsampled_interpolation = nk.signal_resample(signal, method="interpolation", sampling_rate=1000, desired_sampling_rate=500)
    >>> downsampled_fft = nk.signal_resample(signal, method="FFT", sampling_rate=1000, desired_sampling_rate=500)
    >>> downsampled_poly = nk.signal_resample(signal, method="poly", sampling_rate=1000, desired_sampling_rate=500)
    >>> downsampled_numpy = nk.signal_resample(signal, method="numpy", sampling_rate=1000, desired_sampling_rate=500)
    >>> downsampled_pandas = nk.signal_resample(signal, method="pandas", sampling_rate=1000, desired_sampling_rate=500)
    >>>
    >>> # Upsample
    >>> upsampled_interpolation = nk.signal_resample(downsampled_interpolation, method="interpolation", sampling_rate=500, desired_sampling_rate=1000)
    >>> upsampled_fft = nk.signal_resample(downsampled_fft, method="FFT", sampling_rate=500, desired_sampling_rate=1000)
    >>> upsampled_poly = nk.signal_resample(downsampled_poly, method="poly", sampling_rate=500, desired_sampling_rate=1000)
    >>> upsampled_numpy = nk.signal_resample(downsampled_numpy, method="numpy", sampling_rate=500, desired_sampling_rate=1000)
    >>> upsampled_pandas = nk.signal_resample(downsampled_pandas, method="pandas", sampling_rate=500, desired_sampling_rate=1000)
    >>>
    >>> # Compare with original
    >>> pd.DataFrame({"Original": signal,
                      "Interpolation": upsampled_interpolation,
                      "FFT": upsampled_fft,
                      "Poly": upsampled_poly,
                      "Numpy": upsampled_numpy,
                      "Pandas": upsampled_pandas}).plot(style='.-')
    >>>
    >>> # Timing benchmarks
    >>> %timeit nk.signal_resample(signal, method="interpolation", sampling_rate=1000, desired_sampling_rate=500)
    >>> %timeit nk.signal_resample(signal, method="FFT", sampling_rate=1000, desired_sampling_rate=500)
    >>> %timeit nk.signal_resample(signal, method="poly", sampling_rate=1000, desired_sampling_rate=500)
    >>> %timeit nk.signal_resample(signal, method="numpy", sampling_rate=1000, desired_sampling_rate=500)
    >>> %timeit nk.signal_resample(signal, method="pandas", sampling_rate=1000, desired_sampling_rate=500)

    See Also
    --------
    scipy.signal.resample_poly, scipy.signal.resample, scipy.ndimage.zoom
    """
    if desired_length is None:
        desired_length = int(np.round(len(signal) * desired_sampling_rate / sampling_rate))

    # Sanity checks
    if len(signal) == desired_length:
        return(signal)

    # Resample
    if method.lower() == "fft":
        resampled = _resample_fft(signal, desired_length)
    elif method.lower() == "poly":
        resampled =  _resample_poly(signal, desired_length)
    elif method.lower() == "numpy":
        resampled =  _resample_numpy(signal, desired_length)
    elif method.lower() == "pandas":
        resampled =  _resample_pandas(signal, desired_length)
    else:
        resampled =  _resample_interpolation(signal, desired_length)

    return(resampled)




# =============================================================================
# Methods
# =============================================================================

def _resample_numpy(signal, desired_length):
    resampled_signal = np.interp(
        np.linspace(0.0, 1.0, desired_length, endpoint=False),  # where to interpolate
        np.linspace(0.0, 1.0, len(signal), endpoint=False),  # known positions
        signal,  # known data points
    )
    return(resampled_signal)


def _resample_interpolation(signal, desired_length):
    resampled_signal = scipy.ndimage.zoom(signal, desired_length/len(signal))
    return(resampled_signal)


def _resample_fft(signal, desired_length):
    resampled_signal = scipy.signal.resample(signal, desired_length)
    return(resampled_signal)


def _resample_poly(signal, desired_length):
    resampled_signal = scipy.signal.resample_poly(signal, desired_length, len(signal))
    return(resampled_signal)


def _resample_pandas(signal, desired_length):
    # Convert to Time Series
    index = pd.date_range('20131212', freq='L', periods=len(signal))
    resampled_signal = pd.Series(signal, index=index)

    # Create resampling factor
    resampling_factor = str(1/(desired_length / len(signal))) + "L"

    # Resample
    resampled_signal = resampled_signal.resample(resampling_factor).bfill().values

    # Sanitize
    resampled_signal = _resample_sanitize(resampled_signal, desired_length)

    return(resampled_signal)


# =============================================================================
# Internals
# =============================================================================

def _resample_sanitize(resampled_signal, desired_length):
    # Adjust extremities
    diff = len(resampled_signal) - desired_length
    if diff < 0:
        resampled_signal = np.concatenate([resampled_signal, np.full(np.abs(diff), resampled_signal[-1])])
    elif diff > 0:
        resampled_signal = resampled_signal[0:desired_length]
    return resampled_signal


# -*- coding: utf-8 -*-
import numpy as np
import pandas as pd


def rsp_findpeaks(rsp_cleaned, method="khodadad2018", outlier_threshold=0.3):
    """Identify extrema in a respiration (RSP) signal.

    Identify inhalation peaks and exhalation troughs in a preprocessed
    respiration signal using different sets of parameters, such as:

    - `Khodadad et al. (2018)
    <https://iopscience.iop.org/article/10.1088/1361-6579/aad7e6/meta>`_
    - `BioSPPy
    <https://github.com/PIA-Group/BioSPPy/blob/master/biosppy/signals/resp.py>`_

    Parameters
    ----------
    rsp_cleaned : list, array or Series
        The cleaned respiration channel as returned by `rsp_clean()`.
    method : str
        The processing pipeline to apply. Can be one of "khodadad2018"
        (default) or "biosppy".
    outlier_threshold : float
        Only applies if method is "khodadad2018". Extrema that have a vertical
        distance smaller than (outlier_threshold * average vertical distance)
        to any direct neighbour are removed as false positive outliers. I.e.,
        outlier_threshold should be a float with positive sign (the default is
        0.3). Larger values of outlier_threshold correspond to more
        conservative thresholds (i.e., more extrema removed as outliers).

    Returns
    -------
    signals : DataFrame
        A DataFrame of same length as the input signal in which occurences of
        inhalation peaks and exhalation troughs are marked as "1" in lists of
        zeros with the same length as `rsp_cleaned`. Accessible with the keys
        "RSP_Peaks" and "RSP_Troughs" respectively.
    info : dict
        A dictionary containing additional information, in this case the
        samples at which inhalation peaks and exhalation troughs occur,
        accessible with the keys "RSP_Peaks", and "RSP_Troughs", respectively.

    See Also
    --------
    rsp_clean, rsp_rate, rsp_amplitude, rsp_process, rsp_plot

    Examples
    --------
    >>> import neurokit2 as nk
    >>>
    >>> rsp = nk.rsp_simulate(duration=90, respiratory_rate=15)
    >>> cleaned = nk.rsp_clean(rsp, sampling_rate=1000)
    >>> signals, info = nk.rsp_findpeaks(cleaned)
    >>> nk.events_plot([info["RSP_Peaks"], info["RSP_Troughs"]], cleaned)
    """
    # Try retrieving correct column.
    if isinstance(rsp_cleaned, pd.DataFrame):
        try:
            rsp_cleaned = rsp_cleaned["RSP_Clean"]
        except NameError:
            try:
                rsp_cleaned = rsp_cleaned["RSP_Raw"]
            except NameError:
                rsp_cleaned = rsp_cleaned["RSP"]

    cleaned = np.array(rsp_cleaned)

    # Find peaks
    method = method.lower()  # remove capitalised letters
    if method in ["khodadad", "khodadad2018"]:
        peaks, troughs = _rsp_findpeaks_khodadad(cleaned, outlier_threshold)
    elif method == "biosppy":
        peaks, troughs = _rsp_findpeaks_biosppy(cleaned)
    else:
        raise ValueError("NeuroKit error: rsp_findpeaks(): 'method' should be "
                         "one of 'khodadad2018' or 'biosppy'.")

    # Prepare output.
    peaks_signal = np.full(len(rsp_cleaned), 0)
    peaks_signal[peaks] = 1
    troughs_signal = np.full(len(rsp_cleaned), 0)
    troughs_signal[troughs] = 1

    signals = pd.DataFrame({"RSP_Peaks": peaks_signal,
                            "RSP_Troughs": troughs_signal})

    info = {"RSP_Peaks": peaks,
            "RSP_Troughs": troughs}
    return signals, info

# =============================================================================
# Methods
# =============================================================================
def _rsp_findpeaks_biosppy(rsp_cleaned):
    return _rsp_findpeaks_khodadad(rsp_cleaned, outlier_threshold=0)


def _rsp_findpeaks_khodadad(rsp_cleaned, outlier_threshold=0.3):

    extrema = _rsp_findpeaks_extrema(rsp_cleaned)
    extrema, amplitudes = _rsp_findpeaks_outliers(rsp_cleaned, extrema,
                                                  outlier_threshold=outlier_threshold)
    peaks, troughs = _rsp_findpeaks_sanitize(extrema, amplitudes)

    return peaks, troughs

# =============================================================================
# Internals
# =============================================================================
def _rsp_findpeaks_extrema(rsp_cleaned):
    # Detect zero crossings (note that these are zero crossings in the raw
    # signal, not in its gradient).
    greater = rsp_cleaned > 0
    smaller = rsp_cleaned < 0
    risex = np.where(np.bitwise_and(smaller[:-1], greater[1:]))[0]
    fallx = np.where(np.bitwise_and(greater[:-1], smaller[1:]))[0]

    if risex[0] < fallx[0]:
        startx = "rise"
    elif fallx[0] < risex[0]:
        startx = "fall"

    allx = np.concatenate((risex, fallx))
    allx.sort(kind="mergesort")

    # Find extrema by searching minima between falling zero crossing and
    # rising zero crossing, and searching maxima between rising zero
    # crossing and falling zero crossing.
    extrema = []
    for i in range(len(allx) - 1):

        # Determine whether to search for minimum or maximum.
        if startx == "rise":
            if (i + 1) % 2 != 0:
                argextreme = np.argmax
            else:
                argextreme = np.argmin
        elif startx == "fall":
            if (i + 1) % 2 != 0:
                argextreme = np.argmin
            else:
                argextreme = np.argmax

        # Get the two zero crossings between which the extreme will be
        # searched.
        beg = allx[i]
        end = allx[i + 1]

        extreme = argextreme(rsp_cleaned[beg:end])
        extrema.append(beg + extreme)

    extrema = np.asarray(extrema)
    return extrema


def _rsp_findpeaks_outliers(rsp_cleaned, extrema, outlier_threshold=0.3):

    # Only consider those extrema that have a minimum vertical distance to
    # their direct neighbor, i.e., define outliers in absolute amplitude
    # difference between neighboring extrema.
    vertical_diff = np.abs(np.diff(rsp_cleaned[extrema]))
    average_diff = np.mean(vertical_diff)
    min_diff = np.where(vertical_diff > (average_diff * outlier_threshold))[0]
    extrema = extrema[min_diff]

    # Make sure that the alternation of peaks and troughs is unbroken. If
    # alternation of sign in extdiffs is broken, remove the extrema that
    # cause the breaks.
    amplitudes = rsp_cleaned[extrema]
    extdiffs = np.sign(np.diff(amplitudes))
    extdiffs = np.add(extdiffs[0:-1], extdiffs[1:])
    removeext = np.where(extdiffs != 0)[0] + 1
    extrema = np.delete(extrema, removeext)
    amplitudes = np.delete(amplitudes, removeext)

    return extrema, amplitudes


def _rsp_findpeaks_sanitize(extrema, amplitudes):
    # To be able to consistently calculate breathing amplitude, make sure that
    # the extrema always start with a trough and end with a peak, since
    # breathing amplitude will be defined as vertical distance between each
    # peak and the preceding trough. Note that this also ensures that the
    # number of peaks and troughs is equal.
    if amplitudes[0] > amplitudes[1]:
        extrema = np.delete(extrema, 0)
    if amplitudes[-1] < amplitudes[-2]:
        extrema = np.delete(extrema, -1)
    peaks = extrema[1::2]
    troughs = extrema[0:-1:2]

    return peaks, troughs



# -*- coding: utf-8 -*-
import pandas as pd
import numpy as np

import matplotlib.cm
import matplotlib.pyplot as plt



def _events_plot(events, color="red", linestyle="--"):
    # Check if events is list of lists
    try:
        len(events[0])
        is_listoflists = True
    except TypeError:
        is_listoflists = False

    if is_listoflists is False:
        # Loop through sublists
        for event in events:
            plt.axvline(event, color=color, linestyle=linestyle)

    else:
        # Convert color and style to list
        if isinstance(color, str):
            color_map = matplotlib.cm.get_cmap('rainbow')
            color = color_map(np.linspace(0, 1, num=len(events)))
        if isinstance(linestyle, str):
            linestyle = np.full(len(events), linestyle)

        # Loop through sublists
        for i, event in enumerate(events):
            for j in events[i]:
                plt.axvline(j, color=color[i], linestyle=linestyle[i])





def events_plot(events, signal=None, show=True, color="red", linestyle="--"):
    """
    Plot events in signal.

    Parameters
    ----------
    events : list, ndarray or dict
        Events onset location. Can also be a list of lists, in which case it will mark them with different colors. If a dict is passed (e.g., from 'events_find()'), will select only the 'Onset' list.
    signal : array or DataFrame
        Signal array (can be a dataframe with many signals).
    show : bool
        If True, will return a plot. If False, will return a DataFrame that can be plotted externally.
    color, linestyle : str
        Arguments passed to matplotlib plotting.

    See Also
    --------
    events_find

    Examples
    ----------
    >>> import numpy as np
    >>> import pandas as pd
    >>> import neurokit2 as nk
    >>>
    >>> nk.events_plot([1, 3, 5])
    >>>
    >>> signal = np.cos(np.linspace(start=0, stop=20, num=1000))
    >>> events = nk.events_find(signal)
    >>> nk.events_plot(events, signal)
    >>>
    >>> events1 = events["Onset"]
    >>> events2 = np.arange(0, 1000, 100)
    >>> nk.events_plot([events1, events2], signal)
    >>>
    >>> signal = np.cos(np.linspace(start=0, stop=70, num=1000))
    >>> events = nk.events_find(signal)
    >>> events = [[i] for i in events['Onset']]
    >>> nk.events_plot(events, signal)
    """

    if isinstance(events, dict):
        events = events["Onset"]

    if signal is None:
        signal = np.full(events[-1]+1, 0)
    if isinstance(signal, pd.DataFrame) is False:
        signal = pd.DataFrame({"Signal": signal})

    # Plot if necessary
    if show:
        signal.plot()
        _events_plot(events, color=color, linestyle=linestyle)

    else:
        signal["Event_Onset"] = 0
        signal.iloc[events] = 1
        return(signal)



# -*- coding: utf-8 -*-
import pandas as pd
import numpy as np
import itertools

from ..signal import signal_binarize










def _events_find(event_channel, threshold="auto", threshold_keep="above"):
    """Internal function
    """
    binary = signal_binarize(event_channel, threshold=threshold)

    if threshold_keep.lower() != 'above':
        binary = np.abs(binary - 1)  # Reverse if events are below

    # Initialize data
    events = {"Onset":[], "Duration":[]}

    index = 0
    for event, group in (itertools.groupby(binary)):
        duration = len(list(group))
        if event == 1:
            events["Onset"].append(index)
            events["Duration"].append(duration)
        index += duration

    # Convert to array
    events["Onset"] = np.array(events["Onset"])
    events["Duration"] = np.array(events["Duration"])
    return(events)













# ==============================================================================
# ==============================================================================
# ==============================================================================
# ==============================================================================
# ==============================================================================
# ==============================================================================
# ==============================================================================
# ==============================================================================
def events_find(event_channel, threshold="auto", threshold_keep="above", start_at=0, end_at=None, duration_min=1, duration_max=None, inter_min=0, discard_first=0, discard_last=0):
    """Find and select events in a continuous signal (e.g., from a photosensor).

    Parameters
    ----------
    event_channel : array or list
        The channel containing the events.
    threshold : str or float
        The threshold value by which to select the events. If "auto", takes the value between the max and the min.
    threshold_keep : str
        "above" or "below", define the events as above or under the threshold. For photosensors, a white screen corresponds usually to higher values. Therefore, if your events are signaled by a black colour, events values are the lower ones, and you should set the cut to "below".
    start_at, end_at : int
        Keep events which onset is after, or before a particular time point.
    duration_min, duration_max : int
        The minimum or maximum duration of an event to be considered as such (in time points).
    inter_min : int
        The minimum duration after an event for the subsequent event to be considered as such (in time points). Useful when spurious consecutive events are created due to very high sampling rate.
    discard_first, discard_last : int
        Discard first or last n events. Useful if the experiment stats or ends with some spurious events. If discard_first=0 and discard_last=0, no first event or last event is removed.

    Returns
    ----------
    dict
        Dict containing two arrays, 'onset' for events onsets and 'duration' for events durations.

    See Also
    --------
    events_plot, events_to_mne

    Example
    ----------
    >>> import numpy as np
    >>> import pandas as pd
    >>> import neurokit2 as nk
    >>>
    >>> signal = np.cos(np.linspace(start=0, stop=20, num=1000))
    >>> events = nk.events_find(signal)
    >>> events
    {'Onset': array([  0, 236, 550, 864]), 'Duration': array([ 79, 157, 157, 136])}
    >>>
    >>> nk.events_plot(signal, events)
    """
    events = _events_find(event_channel, threshold=threshold, threshold_keep=threshold_keep)

    # Warning when no events detected
    if len(events["Onset"]) == 0:
        print("NeuroKit warning: events_find(): No events found. Check your event_channel or adjust 'threhsold' or 'keep' arguments.")
        return(events)

    # Remove based on duration
    to_keep = np.full(len(events["Onset"]), True)
    to_keep[events["Duration"] < duration_min] = False
    if duration_max is not None:
        to_keep[events["Duration"] > duration_max] = False
    events["Onset"] = events["Onset"][to_keep]
    events["Duration"] = events["Duration"][to_keep]

    # Remove based on index
    if start_at > 0:
        events["Onset"] = events["Onset"][events["Onset"] >= start_at]
        events["Duration"] = events["Duration"][events["Onset"] >= start_at]
    if end_at is not None:
        events["Onset"] = events["Onset"][events["Onset"] <= end_at]
        events["Duration"] = events["Duration"][events["Onset"] <= end_at]

    # Remove based on interval min
    if inter_min > 0:
        inter = np.diff(events["Onset"])
        events["Onset"] = np.concatenate([events["Onset"][0:1], events["Onset"][1::][inter >= inter_min]])
        events["Duration"] = np.concatenate([events["Duration"][0:1], events["Duration"][1::][inter >= inter_min]])

    # Remove first and last n
    if discard_first > 0:
        events["Onset"] = events["Onset"][discard_first:]
        events["Duration"] = events["Duration"][discard_first:]
    if discard_last > 0:
        events["Onset"] = events["Onset"][0:-1*discard_last]
        events["Duration"] = events["Duration"][0:-1*discard_last]

    return(events)



# -*- coding: utf-8 -*-
import pandas as pd
import numpy as np
import scipy
import math

from ..signal import signal_resample
from ..signal import signal_distord


def ecg_simulate(duration=10, length=None, sampling_rate=1000, noise=0.01,
                 heart_rate=70, method="ecgsyn", random_state=42):
    """Simulate an ECG/EKG signal

    Generate an artificial (synthetic) ECG signal of a given duration and sampling rate using either the ECGSYN dynamical model (McSharry et al., 2003) or a simpler model based on Daubechies wavelets to roughly approximate cardiac cycles.

    Parameters
    ----------
    duration : int
        Desired recording length in seconds.
    sampling_rate, length : int
        The desired sampling rate (in Hz, i.e., samples/second) or the desired length of the signal (in samples).
    noise : float
        Noise level (amplitude of the laplace noise).
    heart_rate : int
        Desired simulated heart rate (in beats per minute).
    method : str
        The model used to generate the signal. Can be 'simple' for a
        simulation based on Daubechies wavelets that roughly approximates
        a single cardiac cycle. If 'ecgsyn' (default), will use an
        advanced model desbribed `McSharry et al. (2003)
        <https://physionet.org/content/ecgsyn/>`_.
    random_state : int
        Seed for the random number generator.



    Returns
    ----------
    array
        Vector containing the ECG signal.

    Examples
    ----------
    >>> import pandas as pd
    >>> import neurokit as nk
    >>>
    >>> ecg1 = nk.ecg_simulate(duration=10, method="simple")
    >>> ecg2 = nk.ecg_simulate(duration=10, method="ecgsyn")
    >>> pd.DataFrame({"ECG_Simple": ecg1,
                      "ECG_Complex": ecg2}).plot(subplots=True)

    See Also
    --------
    rsp_simulate, eda_simulate, ppg_simulate, emg_simulate


    References
    -----------
    - McSharry, P. E., Clifford, G. D., Tarassenko, L., & Smith, L. A. (2003). A dynamical model for generating synthetic electrocardiogram signals. IEEE transactions on biomedical engineering, 50(3), 289-294.
    - https://github.com/diarmaidocualain/ecg_simulation
    """
    # Generate number of samples automatically if length is unspecified
    if length is None:
        length = duration * sampling_rate
    if duration is None:
        duration = length / sampling_rate

    # Run appropriate method
    if method.lower() in ["simple", "daubechies"]:
        ecg = _ecg_simulate_daubechies(duration=duration,
                                       length=length,
                                       sampling_rate=sampling_rate,
                                       noise=noise,
                                       heart_rate=heart_rate,
                                       random_state=random_state)
    else:
        approx_number_beats = int(np.round(duration * (heart_rate / 60)))
        ecg = _ecg_simulate_ecgsyn(sfecg=sampling_rate,
                                   N=approx_number_beats,
                                   Anoise=0,
                                   hrmean=heart_rate,
                                   hrstd=1,
                                   lfhfratio=0.5,
                                   sfint=sampling_rate,
                                   ti=(-70, -15, 0, 15, 100),
                                   ai=(1.2, -5, 30, -7.5, 0.75),
                                   bi=(0.25, 0.1, 0.1, 0.1, 0.4),
                                   random_state=random_state)
        # Cut to match expected length
        ecg = ecg[0:length]

    # Add random noise
    if noise > 0:
        ecg = signal_distord(ecg,
                             sampling_rate=sampling_rate,
                             noise_amplitude=noise,
                             noise_frequency=[5, 10, 100],
                             noise_shape="laplace")

    return(ecg)









# =============================================================================
# Daubechies
# =============================================================================
def _ecg_simulate_daubechies(duration=10, length=None, sampling_rate=1000, noise=0.01,
                             heart_rate=70, random_state=42):
    """Generate an artificial (synthetic) ECG signal of a given duration and sampling rate.
    It uses a 'Daubechies' wavelet that roughly approximates a single cardiac cycle.
    This function is based on `this script <https://github.com/diarmaidocualain/ecg_simulation>`_.
    """

    # Seed the random generator for reproducible results
    np.random.seed(random_state)

    # The "Daubechies" wavelet is a rough approximation to a real, single, cardiac cycle
    cardiac = scipy.signal.wavelets.daub(10)

    # Add the gap after the pqrst when the heart is resting.
    cardiac = np.concatenate([cardiac, np.zeros(10)])

    # Caculate the number of beats in capture time period
    num_heart_beats = int(duration * heart_rate / 60)

    # Concatenate together the number of heart beats needed
    ecg = np.tile(cardiac , num_heart_beats)

    # Add random (gaussian distributed) noise
    ecg += np.random.normal(0, noise, len(ecg))

    # Resample
    ecg = signal_resample(ecg,
                          sampling_rate=int(len(ecg)/10),
                          desired_length=length,
                          desired_sampling_rate=sampling_rate)

    return(ecg)



# =============================================================================
# ECGSYN
# =============================================================================
def _ecg_simulate_ecgsyn(sfecg=256, N=256, Anoise=0, hrmean=60, hrstd=1, lfhfratio=0.5, sfint=512,
                         ti=(-70, -15, 0, 15, 100), ai=(1.2, -5, 30, -7.5, 0.75), bi=(0.25, 0.1, 0.1, 0.1, 0.4),
                         random_state=42):
    """
    This function is a python translation of the matlab script by `McSharry & Clifford (2013) <https://physionet.org/content/ecgsyn>`_.

    Parameters
    ----------
    % Operation uses the following parameters (default values in []s):
    % sfecg: ECG sampling frequency [256 Hertz]
    % N: approximate number of heart beats [256]
    % Anoise: Additive uniformly distributed measurement noise [0 mV]
    % hrmean: Mean heart rate [60 beats per minute]
    % hrstd: Standard deviation of heart rate [1 beat per minute]
    % lfhfratio: LF/HF ratio [0.5]
    % sfint: Internal sampling frequency [256 Hertz]
    % Order of extrema: (P Q R S T)
    % ti = angles of extrema (in degrees)
    % ai = z-position of extrema
    % bi = Gaussian width of peaks

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>>
    >>> s = _ecg_simulate_ecgsynth()
    >>> x = np.linspace(0, len(s)-1, len(s))
    >>> num_points = 4000
    >>>
    >>> num_points = min(num_points, len(s))
    >>> plt.plot(x[:num_points], s[:num_points])
    >>> plt.show()
    """
    # Seed the random generator for reproducible results
    np.random.seed(random_state)


    if not isinstance(ti, np.ndarray):
        ti = np.array(ti)
    if not isinstance(ai, np.ndarray):
        ai = np.array(ai)
    if not isinstance(bi, np.ndarray):
        bi = np.array(bi)

    ti = ti*np.pi/180

    # Adjust extrema parameters for mean heart rate
    hrfact = np.sqrt(hrmean/60)
    hrfact2 = np.sqrt(hrfact)
    bi = hrfact * bi
    ti = np.array([hrfact2, hrfact, 1, hrfact, hrfact2])*ti

    # Check that sfint is an integer multiple of sfecg
    q = np.round(sfint/sfecg)
    qd = sfint/sfecg
    if q != qd:
        raise ValueError('Internal sampling frequency (sfint) must be an integer multiple of the ECG sampling frequency (sfecg). Your current choices are: sfecg = ' + str(sfecg) + ' and sfint = ' + str(sfint) + '.')

    # Define frequency parameters for rr process
    # flo and fhi correspond to the Mayer waves and respiratory rate respectively
    flo = 0.1
    fhi = 0.25
    flostd = 0.01
    fhistd = 0.01
    fid = 1

    # Calculate time scales for rr and total output
    sfrr = 1
    trr = 1/sfrr
    tstep = 1/sfecg
    rrmean = 60/hrmean
    n = 2**(np.ceil(np.log2(N*rrmean/trr)))

    rr0 = _ecg_simulate_rrprocess(flo, fhi, flostd, fhistd, lfhfratio, hrmean, hrstd, sfrr, n)

    # Upsample rr time series from 1 Hz to sfint Hz
    rr = signal_resample(rr0, sampling_rate=1, desired_sampling_rate=sfint)

    # Make the rrn time series
    dt = 1/sfint
    rrn = np.zeros(len(rr))
    tecg = 0
    i = 0
    while i < len(rr):
        tecg = tecg+rr[i]
        ip = int(np.round(tecg/dt))
        rrn[i:ip] = rr[i]
        i = ip
    Nt = ip

    # Integrate system using fourth order Runge-Kutta
    x0 = np.array([1, 0, 0.04])

    # tspan is a tuple of (min, max) which defines the lower and upper bound of t in ODE
    # t_eval is the list of desired t points for ODE
    # in Matlab, ode45 can accepts both tspan and t_eval in one argument
    Tspan = [0, (Nt-1)*dt]
    t_eval = np.linspace(0, (Nt-1)*dt, Nt)

    # as passing extra arguments to derivative function is not supported yet in solve_ivp
    # lambda function is used to serve the purpose
    result = scipy.integrate.solve_ivp(lambda t, x: _ecg_simulate_derivsecgsyn(t, x, rrn, ti, sfint, ai, bi),
                                       Tspan,
                                       x0,
                                       t_eval=t_eval
                                       )
    T = result.t
    X0 = result.y

    # downsample to required sfecg
    X = X0[:, np.arange(0, X0.shape[1], q).astype(int)]

    # Scale signal to lie between -0.4 and 1.2 mV
    z = X[2, :].copy()
    zmin = np.min(z)
    zmax = np.max(z)
    zrange = zmax - zmin
    z = (z - zmin)*1.6/zrange - 0.4

    # include additive uniformly distributed measurement noise
    eta = 2*np.random.uniform(len(z))-1
    s = z + Anoise*eta

    return s


def _ecg_simulate_derivsecgsyn(t, x, rr, ti, sfint, ai, bi):
    xi = np.cos(ti)
    yi = np.sin(ti)

    ta = math.atan2(x[1], x[0])
    r0 = 1
    a0 = 1.0 - np.sqrt(x[0] ** 2 + x[1] ** 2) / r0

    ip = np.floor(t * sfint).astype(int)
    w0 = 2*np.pi/rr[min(ip, len(rr)-1)]
    # w0 = 2*np.pi/rr[ip[ip <= np.max(rr)]]

    fresp = 0.25
    zbase = 0.005 * np.sin(2*np.pi*fresp*t)

    dx1dt = a0 * x[0] - w0 * x[1]
    dx2dt = a0 * x[1] + w0 * x[0]

    # matlab rem and numpy rem are different
    # dti = np.remainder(ta - ti, 2*np.pi)
    dti = (ta - ti) - np.round((ta - ti) / 2 / np.pi) * 2 * np.pi
    dx3dt = -np.sum(ai * dti * np.exp(-0.5 * (dti / bi) ** 2)) - 1 * (x[2] - zbase)

    dxdt = np.array([dx1dt, dx2dt, dx3dt])
    return dxdt


def _ecg_simulate_rrprocess(flo=0.1, fhi=0.25, flostd=0.01, fhistd=0.01, lfhfratio=0.5, hrmean=60, hrstd=1, sfrr=1, n=256):
    w1 = 2*np.pi*flo
    w2 = 2*np.pi*fhi
    c1 = 2*np.pi*flostd
    c2 = 2*np.pi*fhistd
    sig2 = 1
    sig1 = lfhfratio
    rrmean = 60/hrmean
    rrstd = 60*hrstd/(hrmean*hrmean)

    df = sfrr/n
    w = np.arange(n)*2*np.pi*df
    dw1 = w-w1
    dw2 = w-w2

    Hw1 = sig1*np.exp(-0.5*(dw1/c1)**2)/np.sqrt(2*np.pi*c1**2)
    Hw2 = sig2*np.exp(-0.5*(dw2/c2)**2)/np.sqrt(2*np.pi*c2**2)
    Hw = Hw1 + Hw2
    Hw0 = np.concatenate((Hw[0:int(n/2)], Hw[int(n/2)-1::-1]))
    Sw = (sfrr/2)*np.sqrt(Hw0)

    ph0 = 2*np.pi*np.random.uniform(size=int(n/2-1))
    ph = np.concatenate([[0], ph0, [0], -np.flipud(ph0)])
    SwC = Sw * np.exp(1j*ph)
    x = (1/n)*np.real(np.fft.ifft(SwC))

    xstd = np.std(x)
    ratio = rrstd/xstd
    rr = rrmean + x*ratio
    return rr



# -*- coding: utf-8 -*-
import pandas as pd

from .ecg_clean import ecg_clean
from .ecg_findpeaks import ecg_findpeaks
from .ecg_rate import ecg_rate


def ecg_process(ecg_signal, sampling_rate=1000, method="neurokit"):
    """"Process an ECG signal.

    Convenience function that automatically processes an ECG signal.

    Parameters
    ----------
    ecg_signal : list, array or Series
        The raw ECG channel.
    sampling_rate : int
        The sampling frequency of `ecg_signal` (in Hz, i.e., samples/second).
        Defaults to 1000.
    method : str
        The processing pipeline to apply. Defaults to "neurokit".

    Returns
    -------
    signals : DataFrame
        A DataFrame of the same length as the `ecg_signal` containing the
        following columns:
        - *"ECG_Raw"*: the raw signal.
        - *"ECG_Clean"*: the cleaned signal.
        - *"ECG_Peaks"*: the R-peaks marked as "1" in a list of zeros.
        - *"ECG_Rate"*: heart rate interpolated between R-peaks.
    info : dict
        A dictionary containing the samples at which the R-peaks occur,
        accessible with the key "ECG_Peaks".

    See Also
    --------
    ecg_clean, ecg_findpeaks, ecg_rate, ecg_plot

    Examples
    --------
    >>> import neurokit2 as nk
    >>>
    >>> ecg = nk.ecg_simulate(duration=15, sampling_rate=1000, heart_rate=80)
    >>> signals, info = nk.ecg_process(ecg, sampling_rate=1000)
    >>> nk.ecg_plot(signals)

    """
    ecg_cleaned = ecg_clean(ecg_signal, sampling_rate=sampling_rate,
                            method=method)

    extrema_signal, info = ecg_findpeaks(ecg_cleaned=ecg_cleaned,
                                         sampling_rate=sampling_rate,
                                         method=method,
                                         show=False)

    rate = ecg_rate(extrema_signal, sampling_rate=sampling_rate)

    signals = pd.DataFrame({"ECG_Raw": ecg_signal,
                            "ECG_Clean": ecg_cleaned})
    signals = pd.concat([signals, extrema_signal, rate], axis=1)
    return signals, info



# -*- coding: utf-8 -*-
import pandas as pd
import numpy as np

import scipy.stats







def _standardize(data, robust=False):

    if robust is False:
        z = (data - np.mean(data, axis=0))/np.std(data, axis=0)
    else:
        z = (data - np.median(data, axis=0))/scipy.stats.median_absolute_deviation(data, axis=0, nan_policy="omit")

    return(z)






def standardize(data, robust=False):
    """Standardization of data

    Performs a standardization of data (Z-scoring), i.e., centering and scaling,
    so that the data is expressed in terms of standard deviation
    (i.e., mean = 0, SD = 1) or Median Absolute Deviance (median = 0, MAD = 1).

    Parameters
    ----------
    data : list, array or Series
        Raw data.
    robust : bool
        If True, centering is done by substracting the median from the
        variables and dividing it by the median absolute deviation (MAD).
        If False, variables are standardized by substracting the mean and
        dividing it by the standard deviation (SD).

    Returns
    ----------
    list, array or Series
        The standardized values.


    Examples
    ----------
    >>> import neurokit2 as nk
    >>>
    >>> nk.standardize([3, 1, 2, 4, 6])
    """
    # Return appropriate type
    if isinstance(data, list):
        data = list(_standardize(np.array(data), robust=robust))
    elif isinstance(data, pd.Series) or isinstance(data, pd.DataFrame):
        z = _standardize(data.values, robust=robust)
        data[:] = z
    else:
        data = _standardize(data, robust=robust)

    return(data)



# -*- coding: utf-8 -*-
import numpy as np

from ..signal import signal_distord


def rsp_simulate(duration=10, length=None, sampling_rate=1000, noise=0.01,
                 respiratory_rate=15, method="breathmetrics", random_state=42):
    """Simulate a respiratory signal.

    Generate an artificial (synthetic) respiratory signal of a given duration
    and rate.

    Parameters
    ----------
    duration : int
        Desired length of duration (s).
    sampling_rate, length : int
        The desired sampling rate (in Hz, i.e., samples/second) or the desired
        length of the signal (in samples).
    noise : float
        Noise level (amplitude of the laplace noise).
    respiratory_rate : float
        Desired number of breath cycles in one minute.
    method : str
        The model used to generate the signal. Can be 'sinusoidal' for a
        simulation based on a trigonometric sine wave that roughly approximates
        a single respiratory cycle. If 'breathmetrics' (default), will use an
        advanced model desbribed `Noto, et al. (2018)
        <https://github.com/zelanolab/breathmetrics/blob/master/simulateRespiratoryData.m>`_.
    random_state : int
        Seed for the random number generator.

    Returns
    -------
    array
        Vector containing the respiratory signal.

    Examples
    --------
    >>> import pandas as pd
    >>> import numpy as np
    >>> import neurokit2 as nk
    >>>
    >>> rsp1 = nk.rsp_simulate(duration=30, method="sinusoidal")
    >>> rsp2 = nk.rsp_simulate(duration=30, method="breathmetrics")
    >>> pd.DataFrame({"RSP_Simple": rsp1,
                      "RSP_Complex": rsp2}).plot(subplots=True)

    References
    ----------
    Noto, T., Zhou, G., Schuele, S., Templer, J., & Zelano, C. (2018).
    Automated analysis of breathing waveforms using BreathMetrics: A
    respiratory signal processing toolbox. Chemical Senses, 43(8), 583–597.
    https://doi.org/10.1093/chemse/bjy045

    See Also
    --------
    rsp_clean, rsp_findpeaks, rsp_rate, rsp_process, rsp_plot
"""

    # Seed the random generator for reproducible results
    np.random.seed(random_state)

    # Generate number of samples automatically if length is unspecified
    if length is None:
        length = duration * sampling_rate

    if method.lower() in ["sinusoidal", "sinus", "simple"]:
        rsp = _rsp_simulate_sinusoidal(duration=duration,
                                       length=length,
                                       sampling_rate=sampling_rate,
                                       respiratory_rate=respiratory_rate)
    else:
        rsp = _rsp_simulate_breathmetrics(duration=duration,
                                          length=length,
                                          sampling_rate=sampling_rate,
                                          respiratory_rate=respiratory_rate)
        rsp = rsp[0:length]

    # Add random noise
    if noise > 0:
        rsp = signal_distord(rsp,
                             sampling_rate=sampling_rate,
                             noise_amplitude=noise,
                             noise_frequency=[5, 10, 100],
                             noise_shape="laplace")
    return rsp









# =============================================================================
# Simple Sinusoidal Model
# =============================================================================
def _rsp_simulate_sinusoidal(duration=10, length=None, sampling_rate=1000,
                             noise=0.01, respiratory_rate=15, amplitude=0.5):
    """
    Generate an artificial (synthetic) respiratory signal by trigonometric sine
    wave that roughly approximates a single respiratory cycle.
    """
    # Generate values along the length of the duration
    x = np.linspace(0, duration, int(length))

    frequency = respiratory_rate / 60

    # Compute the value of sine computed by the following trigonometric
    # function
    rsp = amplitude*np.sin(2*np.pi*x*frequency)

    return rsp





# =============================================================================
# BreathMetrics Model
# =============================================================================
def _rsp_simulate_breathmetrics_original(nCycles=100,
                                         sampling_rate=1000,
                                         breathing_rate=0.25,
                                         average_amplitude=0.5,
                                         amplitude_variance=0.1,
                                         phase_variance=0.1,
                                         inhale_pause_percent=0.3,
                                         inhale_pause_avgLength=0.2,
                                         inhale_pauseLength_variance=0.5,
                                         exhale_pause_percent=0.3,
                                         exhale_pause_avgLength=0.2,
                                         exhale_pauseLength_variance=0.5,
                                         pause_amplitude=0.1,
                                         pause_amplitude_variance=0.2,
                                         signal_noise=0.1):
    """
    Simulates a recording of human airflow data by appending individually
    constructed sin waves and pauses in sequence. This is translated from the
    matlab code available `here <https://github.com/zelanolab/breathmetrics/blob/master/simulateRespiratoryData.m>`_
    by Noto, et al. (2018).

    Parameters
    ----------
    nCycles : int or float
        number of breathing cycles to simulate.
    sampling_rate :
        sampling rate.
    breathing_rate :
        average breathing rate.
    average_amplitude :
        average amplitude of inhales and exhales.
    amplitude_variance:
        variance in respiratory amplitudes.
    phase_variance:
        variance in duration of individual breaths.
    inhale_pause_percent :
        percent of inhales followed by a pause.
    inhale_pause_avgLength :
        average length of inhale pauses.
    inhale_pauseLength_variance :
        variance in inhale pause length.
    exhale_pause_percent :
        percent of exhales followed by a pause.
    exhale_pause_avgLength :
        average length of exhale pauses.
    exhale_pauseLength_variance :
        variance in exhale pause length.
    pause_amplitude :
        noise amplitude of pauses.
    pause_amplitude_variance :
        variance in pause noise.
    signal_noise :
        percent of noise saturation in the simulated signal.
    Returns
    ----------
    """
    # Define additional parameters
    sample_phase = sampling_rate / breathing_rate
    inhale_pause_phase = np.round(inhale_pause_avgLength *
                                  sample_phase).astype(int)
    exhale_pause_phase = np.round(exhale_pause_avgLength *
                                  sample_phase).astype(int)

    # Normalize variance by average breath amplitude
    amplitude_variance_normed = average_amplitude * amplitude_variance
    amplitudes_with_noise = np.random.randn(
        nCycles) * amplitude_variance_normed + average_amplitude
    amplitudes_with_noise[amplitudes_with_noise < 0] = 0

    # Normalize phase by average breath length
    phase_variance_normed = phase_variance * sample_phase
    phases_with_noise = np.round(
        np.random.randn(nCycles) * phase_variance_normed +
        sample_phase).astype(int)
    phases_with_noise[phases_with_noise < 0] = 0

    # Normalize pause lengths by phase and variation
    inhale_pauseLength_variance_normed = inhale_pause_phase * inhale_pauseLength_variance
    inhale_pauseLengths_with_noise = np.round(
        np.random.randn(nCycles) * inhale_pauseLength_variance_normed +
        inhale_pause_phase).astype(int)
    inhale_pauseLengths_with_noise[inhale_pauseLengths_with_noise < 0] = 0
    exhale_pauseLength_variance_normed = exhale_pause_phase * exhale_pauseLength_variance
    exhale_pauseLengths_with_noise = np.round(np.random.randn(nCycles) * exhale_pauseLength_variance_normed + inhale_pause_phase).astype(int)

    # why inhale pause phase?
    exhale_pauseLengths_with_noise[exhale_pauseLengths_with_noise < 0] = 0

    # Normalize pause amplitudes
    pause_amplitude_variance_normed = pause_amplitude * pause_amplitude_variance

    # Initialize empty vector to fill with simulated data
    simulated_respiration = []

    # Initialize parameters to save
    inhale_onsets = np.zeros(nCycles)
    exhale_onsets = np.zeros(nCycles)

    inhale_pause_onsets = np.zeros(nCycles)
    exhale_pause_onsets = np.zeros(nCycles)

    inhale_lengths = np.zeros(nCycles)
    inhale_pauseLengths = np.zeros(nCycles)
    exhale_lengths = np.zeros(nCycles)
    exhale_pauseLengths = np.zeros(nCycles)

    inhale_peaks = np.zeros(nCycles)
    exhale_troughs = np.zeros(nCycles)

    i = 1
    for c in range(nCycles):
        # Determine length of inhale pause for this cycle
        if np.random.rand() < inhale_pause_percent:
            this_inhale_pauseLength = inhale_pauseLengths_with_noise[c]
            this_inhale_pause = np.random.randn(
                this_inhale_pauseLength) * pause_amplitude_variance_normed
            this_inhale_pause[this_inhale_pause < 0] = 0
        else:
            this_inhale_pauseLength = 0
            this_inhale_pause = []

        # Determine length of exhale pause for this cycle
        if np.random.rand() < exhale_pause_percent:
            this_exhale_pauseLength = exhale_pauseLengths_with_noise[c]
            this_exhale_pause = np.random.randn(
                this_exhale_pauseLength) * pause_amplitude_variance_normed
            this_exhale_pause[this_exhale_pause < 0] = 0
        else:
            this_exhale_pauseLength = 0
            this_exhale_pause = []

        # Determine length of inhale and exhale for this cycle to main
        # breathing rate
        cycle_length = phases_with_noise[c] - (this_inhale_pauseLength +
                                               this_exhale_pauseLength)

        # If pauses are longer than the time alloted for this breath, set them
        # to 0 so a real breath can be simulated. This will deviate the
        # statistics from those initialized but is unavaoidable at the current
        # state
        if (cycle_length <= 0) or (cycle_length < min(phases_with_noise) / 4):
            this_inhale_pauseLength = 0
            this_inhale_pause = []
            this_exhale_pauseLength = 0
            this_exhale_pause = []
            cycle_length = phases_with_noise[c] - (this_inhale_pauseLength +
                                                   this_exhale_pauseLength)

        # Compute inhale and exhale for this cycle
        this_cycle = np.sin(np.linspace(
            0, 2 * np.pi, cycle_length)) * amplitudes_with_noise[c]
        half_cycle = np.round(len(this_cycle) / 2).astype(int)
        this_inhale = this_cycle[0:half_cycle]
        this_inhale_length = len(this_inhale)
        this_exhale = this_cycle[half_cycle:]
        this_exhale_length = len(this_exhale)

        # Save parameters for checking
        inhale_lengths[c] = this_inhale_length
        inhale_pauseLengths[c] = this_inhale_pauseLength
        exhale_lengths[c] = this_exhale_length
        exhale_pauseLengths[c] = this_exhale_pauseLength
        inhale_onsets[c] = i
        exhale_onsets[c] = i + this_inhale_length + this_inhale_pauseLength

        if len(this_inhale_pause) > 0:
            inhale_pause_onsets[c] = i + this_inhale_length
        else:
            inhale_pause_onsets[c] = np.nan

        if len(this_exhale_pause) > 0:
            exhale_pause_onsets[
                c] = i + this_inhale_length + this_inhale_pauseLength + this_exhale_length
        else:
            exhale_pause_onsets[c] = np.nan

        # Compose breath from parameters
        this_breath = np.hstack(
            [this_inhale, this_inhale_pause, this_exhale, this_exhale_pause])

        # Compute max flow for inhale and exhale for this breath
        max_ID = np.argmax(this_breath)
        min_ID = np.argmin(this_breath)
        inhale_peaks[c] = i + max_ID
        exhale_troughs[c] = i + min_ID

        # Append breath to simulated resperation vector
        simulated_respiration = np.hstack([simulated_respiration, this_breath])
        i = i + len(this_breath) - 1

    if signal_noise == 0:
        signal_noise = 0.0001

    noise_vector = np.random.rand(
        *simulated_respiration.shape) * average_amplitude
    simulated_respiration = simulated_respiration * (
        1 - signal_noise) + noise_vector * signal_noise
    raw_features = {
        'Inhale Onsets': inhale_onsets,
        'Exhale Onsets': exhale_onsets,
        'Inhale Pause Onsets': inhale_pause_onsets,
        'Exhale Pause Onsets': exhale_pause_onsets,
        'Inhale Lengths': inhale_lengths / sampling_rate,
        'Inhale Pause Lengths': inhale_pauseLengths / sampling_rate,
        'Exhale Lengths': exhale_lengths / sampling_rate,
        'Exhale Pause Lengths': exhale_pauseLengths / sampling_rate,
        'Inhale Peaks': inhale_peaks,
        'Exhale Troughs': exhale_troughs
    }
    if len(inhale_pauseLengths[inhale_pauseLengths > 0]) > 0:
        avg_inhale_pauseLength = np.mean(
            inhale_pauseLengths[inhale_pauseLengths > 0])
    else:
        avg_inhale_pauseLength = 0

    if len(exhale_pauseLengths[exhale_pauseLengths > 0]) > 0:
        avg_exhale_pauseLength = np.mean(
            exhale_pauseLengths[exhale_pauseLengths > 0])
    else:
        avg_exhale_pauseLength = 0

    estimated_breathing_rate = (
        1 / np.mean(np.diff(inhale_onsets))) * sampling_rate
    feature_stats = {
        'Breathing Rate': estimated_breathing_rate,
        'Average Inhale Length': np.mean(inhale_lengths / sampling_rate),
        'Average Inhale Pause Length': avg_inhale_pauseLength / sampling_rate,
        'Average Exhale Length': np.mean(exhale_lengths / sampling_rate),
        'Average Exhale Pause Length': avg_exhale_pauseLength / sampling_rate
    }

    return simulated_respiration, raw_features, feature_stats






def _rsp_simulate_breathmetrics(duration=10, length=None, sampling_rate=1000,
                                respiratory_rate=15):
    """
    """
    n_cycles = int(respiratory_rate / 60 * duration)

    # Loop until it doesn't fail
    rsp = False
    while rsp is False:
        # Generate a longer than necessary signal so it won't be shorter
        rsp, raw_features, feature_stats = _rsp_simulate_breathmetrics_original(
                nCycles=int(n_cycles * 1.5),
                sampling_rate=sampling_rate,
                breathing_rate=respiratory_rate/60,
                signal_noise=0)
    return rsp



import numpy as np
import pandas as pd
import neurokit2 as nk

import nolds

from pyentrp import entropy as pyentrp

"""
For the testing of complexity, we test our implementations against existing and established ones.
However, as some of these other implementations are not really packaged in a way
that we can easily import them. Thus, we directly copied their content in this file
(below the tests).
"""


# =============================================================================
# Complexity
# =============================================================================


def test_complexity():

    signal = np.cos(np.linspace(start=0, stop=30, num=100))


    # Shannon
    assert np.allclose(nk.entropy_shannon(signal) - pyentrp.shannon_entropy(signal), 0)


    # Approximate
    assert np.allclose(nk.entropy_approximate(signal), 0.17364897858477146)
    assert np.allclose(nk.entropy_approximate(signal, 2, 0.2*np.std(signal, ddof=1)) - entropy_app_entropy(signal, 2), 0)

    assert nk.entropy_approximate(signal, 2, 0.2*np.std(signal, ddof=1)) != pyeeg_ap_entropy(signal, 2, 0.2*np.std(signal, ddof=1))


    # Sample
    assert np.allclose(nk.entropy_sample(signal, 2, 0.2*np.std(signal, ddof=1)) - entropy_sample_entropy(signal, 2), 0)
    assert np.allclose(nk.entropy_sample(signal, 2, 0.2) - nolds.sampen(signal, 2, 0.2), 0)
    assert np.allclose(nk.entropy_sample(signal, 2, 0.2) - entro_py_sampen(signal, 2, 0.2, scale=False), 0)
    assert np.allclose(nk.entropy_sample(signal, 2, 0.2) - pyeeg_samp_entropy(signal, 2, 0.2), 0)

    assert nk.entropy_sample(signal, 2, 0.2) != pyentrp.sample_entropy(signal, 2, 0.2)[1]



    # Fuzzy
    assert np.allclose(nk.entropy_fuzzy(signal, 2, 0.2, 1) - entro_py_fuzzyen(signal, 2, 0.2, 1, scale=False), 0)




# =============================================================================
# Pyeeg
# =============================================================================


def pyeeg_embed_seq(time_series, tau, embedding_dimension):
    if not type(time_series) == np.ndarray:
        typed_time_series = np.asarray(time_series)
    else:
        typed_time_series = time_series

    shape = (
        typed_time_series.size - tau * (embedding_dimension - 1),
        embedding_dimension
    )

    strides = (typed_time_series.itemsize, tau * typed_time_series.itemsize)

    return np.lib.stride_tricks.as_strided(
        typed_time_series,
        shape=shape,
        strides=strides
    )




def pyeeg_bin_power(X, Band, Fs):
    C = np.fft.fft(X)
    C = abs(C)
    Power = np.zeros(len(Band) - 1)
    for Freq_Index in range(0, len(Band) - 1):
        Freq = float(Band[Freq_Index])
        Next_Freq = float(Band[Freq_Index + 1])
        Power[Freq_Index] = sum(
            C[int(np.floor(Freq / Fs * len(X))):
                int(np.floor(Next_Freq / Fs * len(X)))]
        )
    Power_Ratio = Power / sum(Power)
    return Power, Power_Ratio





def pyeeg_ap_entropy(X, M, R):
    N = len(X)

    Em = pyeeg_embed_seq(X, 1, M)
    A = np.tile(Em, (len(Em), 1, 1))
    B = np.transpose(A, [1, 0, 2])
    D = np.abs(A - B)  # D[i,j,k] = |Em[i][k] - Em[j][k]|
    InRange = np.max(D, axis=2) <= R

    # Probability that random M-sequences are in range
    Cm = InRange.mean(axis=0)

    # M+1-sequences in range if M-sequences are in range & last values are close
    Dp = np.abs(
        np.tile(X[M:], (N - M, 1)) - np.tile(X[M:], (N - M, 1)).T
    )

    Cmp = np.logical_and(Dp <= R, InRange[:-1, :-1]).mean(axis=0)

    Phi_m, Phi_mp = np.sum(np.log(Cm)), np.sum(np.log(Cmp))

    Ap_En = (Phi_m - Phi_mp) / (N - M)

    return Ap_En


def pyeeg_samp_entropy(X, M, R):
    N = len(X)

    Em = pyeeg_embed_seq(X, 1, M)[:-1]
    A = np.tile(Em, (len(Em), 1, 1))
    B = np.transpose(A, [1, 0, 2])
    D = np.abs(A - B)  # D[i,j,k] = |Em[i][k] - Em[j][k]|
    InRange = np.max(D, axis=2) <= R
    np.fill_diagonal(InRange, 0)  # Don't count self-matches

    Cm = InRange.sum(axis=0)  # Probability that random M-sequences are in range
    Dp = np.abs(
        np.tile(X[M:], (N - M, 1)) - np.tile(X[M:], (N - M, 1)).T
    )

    Cmp = np.logical_and(Dp <= R, InRange).sum(axis=0)

    # Avoid taking log(0)
    Samp_En = np.log(np.sum(Cm + 1e-100) / np.sum(Cmp + 1e-100))

    return Samp_En

# =============================================================================
# Entropy
# =============================================================================


from math import log, floor
from sklearn.neighbors import KDTree

def entropy_embed(x, order=3, delay=1):
    N = len(x)
    if order * delay > N:
        raise ValueError("Error: order * delay should be lower than x.size")
    if delay < 1:
        raise ValueError("Delay has to be at least 1.")
    if order < 2:
        raise ValueError("Order has to be at least 2.")
    Y = np.zeros((order, N - (order - 1) * delay))
    for i in range(order):
        Y[i] = x[i * delay:i * delay + Y.shape[1]]
    return Y.T



def entropy_app_samp_entropy(x, order, metric='chebyshev', approximate=True):
    _all_metrics = KDTree.valid_metrics
    if metric not in _all_metrics:
        raise ValueError('The given metric (%s) is not valid. The valid '
                         'metric names are: %s' % (metric, _all_metrics))
    phi = np.zeros(2)
    r = 0.2 * np.std(x, axis=-1, ddof=1)

    # compute phi(order, r)
    _emb_data1 = entropy_embed(x, order, 1)
    if approximate:
        emb_data1 = _emb_data1
    else:
        emb_data1 = _emb_data1[:-1]
    count1 = KDTree(emb_data1, metric=metric).query_radius(emb_data1, r,
                                                           count_only=True
                                                           ).astype(np.float64)
    # compute phi(order + 1, r)
    emb_data2 = entropy_embed(x, order + 1, 1)
    count2 = KDTree(emb_data2, metric=metric).query_radius(emb_data2, r,
                                                           count_only=True
                                                           ).astype(np.float64)
    if approximate:
        phi[0] = np.mean(np.log(count1 / emb_data1.shape[0]))
        phi[1] = np.mean(np.log(count2 / emb_data2.shape[0]))
    else:
        phi[0] = np.mean((count1 - 1) / (emb_data1.shape[0] - 1))
        phi[1] = np.mean((count2 - 1) / (emb_data2.shape[0] - 1))
    return phi




def entropy_app_entropy(x, order=2, metric='chebyshev'):
    phi = entropy_app_samp_entropy(x, order=order, metric=metric, approximate=True)
    return np.subtract(phi[0], phi[1])



def entropy_sample_entropy(x, order=2, metric='chebyshev'):
    x = np.asarray(x, dtype=np.float64)
    phi = entropy_app_samp_entropy(x, order=order, metric=metric,
                            approximate=False)
    return -np.log(np.divide(phi[1], phi[0]))



# =============================================================================
# entro-py
# =============================================================================

def entro_py_sampen(x, dim, r, scale=True):
    return entro_py_entropy(x, dim, r, scale=scale)


def entro_py_cross_sampen(x1, x2, dim, r, scale=True):
    return entro_py_entropy([x1, x2], dim, r, scale)


def entro_py_fuzzyen(x, dim, r, n, scale=True):
    return entro_py_entropy(x, dim, r, n=n, scale=scale, remove_baseline=True)


def entro_py_cross_fuzzyen(x1, x2, dim, r, n, scale=True):
    return entro_py_entropy([x1, x2], dim, r, n, scale=scale, remove_baseline=True)


def entro_py_pattern_mat(x, m):
    x = np.asarray(x).ravel()
    if m == 1:
        return x
    else:
        N = len(x)
        patterns = np.zeros((m, N-m+1))
        for i in range(m):
            patterns[i, :] = x[i:N-m+i+1]
        return patterns


def entro_py_entropy(x, dim, r, n=1, scale=True, remove_baseline=False):
    fuzzy = True if remove_baseline else False
    cross = True if type(x) == list else False
    N = len(x[0]) if cross else len(x)

    if scale:
        if cross:
            x = [entro_py_scale(np.copy(x[0])), entro_py_scale(np.copy(x[1]))]
        else:
            x = entro_py_scale(np.copy(x))

    phi = [0, 0]  # phi(m), phi(m+1)
    for j in [0, 1]:
        m = dim + j
        npat = N-dim  # https://github.com/ixjlyons/entro-py/pull/2/files
        if cross:
#            patterns = [entro_py_pattern_mat(x[0], m), entro_py_pattern_mat(x[1], m)]
            patterns = [entro_py_pattern_mat(x[0], m)[:, :npat], entro_py_pattern_mat(x[1], m)[:, :npat]]  # https://github.com/ixjlyons/entro-py/pull/2/files
        else:
#            patterns = entro_py_pattern_mat(x, m)
            patterns = entro_py_pattern_mat(x, m)[:, :npat]

        if remove_baseline:
            if cross:
                patterns[0] = entro_py_remove_baseline(patterns[0], axis=0)
                patterns[1] = entro_py_remove_baseline(patterns[1], axis=0)
            else:
                patterns = entro_py_remove_baseline(patterns, axis=0)

#        count = np.zeros(N-m)  # https://github.com/ixjlyons/entro-py/pull/2/files
#        for i in range(N-m):  # https://github.com/ixjlyons/entro-py/pull/2/files
        count = np.zeros(npat)
        for i in range(npat):
            if cross:
                if m == 1:
                    sub = patterns[1][i]
                else:
                    sub = patterns[1][:, [i]]
                dist = np.max(np.abs(patterns[0] - sub), axis=0)
            else:
                if m == 1:
                    sub = patterns[i]
                else:
                    sub = patterns[:, [i]]
                dist = np.max(np.abs(patterns - sub), axis=0)

            if fuzzy:
                sim = np.exp(-np.power(dist, n) / r)
            else:
                sim = dist < r

            count[i] = np.sum(sim) - 1

#        phi[j] = np.mean(count) / (N-m-1)
        phi[j] = np.mean(count) / (N-dim-1)  # https://github.com/ixjlyons/entro-py/pull/2/files

    return np.log(phi[0] / phi[1])


def entro_py_scale(x, axis=None):
    x = entro_py_remove_baseline(x, axis=axis)
    x /= np.std(x, ddof=1, axis=axis, keepdims=True)
    return x


def entro_py_remove_baseline(x, axis=None):
    x -= np.mean(x, axis=axis, keepdims=True)
    return x


# -*- coding: utf-8 -*-
import numpy as np
import sklearn.neighbors

from .utils_embed import _embed




def _get_embedded(signal, order, r, metric='chebyshev', approximate=True):
    embedded = _embed(signal, order, 1)
    if approximate is False:
        embedded = embedded[:-1]
    count = sklearn.neighbors.KDTree(embedded, metric=metric).query_radius(embedded, r, count_only=True).astype(np.float64)
    return(embedded, count)






def _phi(signal, order, r="default", metric='chebyshev', approximate=True):
    """Common internal for `entropy_approximate` and `entropy_sample`.

    Adapted from `EntroPy <https://github.com/raphaelvallat/entropy>`_, check it out!
    """
    # Sanity check
    if metric not in sklearn.neighbors.KDTree.valid_metrics:
        raise ValueError("NeuroKit error: _entropy_approximate_and_sample(): The given metric (%s) is not valid. The valid metric names are: %s" % (metric, sklearn.neighbors.KDTree.valid_metrics))

    # Initialize phi
    phi = np.zeros(2)

    # compute phi(order, r)
    embedded1, count1 = _get_embedded(signal, order, r, metric=metric, approximate=approximate)
    # compute phi(order + 1, r)
    embedded2, count2 = _get_embedded(signal, order + 1, r, metric=metric, approximate=True)

    if approximate is True:
        phi[0] = np.mean(np.log(count1 / embedded1.shape[0]))
        phi[1] = np.mean(np.log(count2 / embedded2.shape[0]))
    else:
        phi[0] = np.mean((count1 - 1) / (embedded1.shape[0] - 1))
        phi[1] = np.mean((count2 - 1) / (embedded2.shape[0] - 1))
    return(phi)



# -*- coding: utf-8 -*-
import numpy as np
import scipy.signal

from ..stats import loess

def signal_smooth(signal, method='convolution', kernel='boxzen', size=10, alpha=0.1):
    """Signal smoothing.

    Signal smoothing can be achieved using either the convolution of a filter kernel with the input
    signal to compute the smoothed signal (Smith, 1997) or a LOESS regression.

    Parameters
    ----------
    signal : list, array or Series
        The signal channel in the form of a vector of values.
    method : str
        Can be one of 'convolution' (default) or 'loess'.
    kernel : str, array
        Only used if `method` is 'convolution'. Type of kernel to use; if array, use directly as the kernel. Can be one
        of 'median', 'boxzen', 'boxcar', 'triang', 'blackman', 'hamming',
        'hann', 'bartlett', 'flattop', 'parzen', 'bohman', 'blackmanharris',
        'nuttall', 'barthann', 'kaiser' (needs beta), 'gaussian' (needs std),
        'general_gaussian' (needs power, width), 'slepian' (needs width) or
        'chebwin' (needs attenuation).
    size : int
        Only used if `method` is 'convolution'. Size of the kernel; ignored if kernel is an array.
    alpha : float
        Only used if `method` is 'loess'. The parameter which controls the degree of smoothing.

    Returns
    -------
    array
        Smoothed signal.


    See Also
    ---------
    loess

    Examples
    --------
    >>> import numpy as np
    >>> import pandas as pd
    >>> import neurokit2 as nk
    >>>
    >>> signal = np.cos(np.linspace(start=0, stop=10, num=1000))
    >>> distorted = nk.signal_distord(signal, noise_amplitude=[0.3, 0.2, 0.1, 0.05], noise_frequency=[5, 10, 50, 500])
    >>>
    >>> size = len(signal)/100
    >>> signals = pd.DataFrame({
            "Raw": distorted,
            "Median": nk.signal_smooth(distorted, kernel='median', size=size-1),
            "BoxZen": nk.signal_smooth(distorted, kernel='boxzen', size=size),
            "Triang": nk.signal_smooth(distorted, kernel='triang', size=size),
            "Blackman": nk.signal_smooth(distorted, kernel='blackman', size=size),
            "Loess_01": nk.signal_smooth(distorted, method='loess', alpha=0.1),
            "Loess_02": nk.signal_smooth(distorted, method='loess', alpha=0.2),
            "Loess_05": nk.signal_smooth(distorted, method='loess', alpha=0.5)})
    >>> signals.plot()
    >>> signals[50:150].plot()  # Magnify

    References
    ----------
    - Smith, S. W. (1997). The scientist and engineer's guide to digital signal
    processing.
    """
    length = len(signal)

    if isinstance(kernel, str) is False:
        raise TypeError("NeuroKit error: signal_smooth(): 'kernel' "
                        "should be a string.")

    # Check length.
    if size > length or size < 1:
        raise TypeError("NeuroKit error: signal_smooth(): 'size' "
                        "should be between 1 and length of the signal.")

    method = method.lower()

    # LOESS
    if method in ["loess", "lowess"]:
        smoothed = loess(signal, alpha=alpha)

    # Convolution
    else:
        if kernel == 'boxzen':
            # hybrid method
            # 1st pass - boxcar kernel
            x = _signal_smoothing(signal, kernel='boxcar', size=size)

            # 2nd pass - parzen kernel
            smoothed = _signal_smoothing(x, kernel='parzen', size=size)

        elif kernel == 'median':
            smoothed = _signal_smoothing_median(signal, size)

        else:
            smoothed = _signal_smoothing(signal, kernel=kernel, size=size)

    return smoothed


# =============================================================================
# Internals
# =============================================================================
def _signal_smoothing_median(signal, size=5):

    # Enforce odd kernel size.
    if size % 2 == 0:
        size += 1

    smoothed = scipy.signal.medfilt(signal, kernel_size=int(size))
    return smoothed


def _signal_smoothing(signal, kernel="boxcar", size=5):

    # Get window.
    size = int(size)
    window = scipy.signal.get_window(kernel, size)
    w = window / window.sum()

    # Extend signal edges to avoid boundary effects.
    x = np.concatenate((signal[0] * np.ones(size), signal,
                        signal[-1] * np.ones(size)))

    # Compute moving average.
    smoothed = np.convolve(w, x, mode='same')
    smoothed = smoothed[size:-size]
    return smoothed



# -*- coding: utf-8 -*-
import os
import bioread

import pandas as pd
import numpy as np

from ..signal import signal_resample

def read_acqknowledge(filename, sampling_rate="max", resample_method="numpy", impute_missing=True):
    """Read and format a BIOPAC's AcqKnowledge file into a pandas' dataframe.

    The function outputs both the dataframe and the sampling rate (encoded within the AcqKnowledge) file.

    Parameters
    ----------
    filename :  str
        Filename (with or without the extension) of a BIOPAC's AcqKnowledge file.
    sampling_rate : int
        Sampling rate (in Hz, i.e., samples/second). Since an AcqKnowledge file can contain signals recorded at different rates, harmonization is necessary in order to convert it to a DataFrame. Thus, if `sampling_rate` is set to 'max' (default), will keep the maximum recorded sampling rate and upsample the channels with lower rate if necessary (using the `signal_resample()` function). If the sampling rate is set to a given value, will resample the signals to the desired value. Note that the value of the sampling rate is outputted along with the data.
    resample_method : str
        Method of resampling (see `signal_resample()`).
    impute_missing : bool
        Sometimes, due to connections issues, the signal has some holes (short periods without signal). If 'impute_missing' is True, will automatically fill the signal interruptions using padding.

    Returns
    ----------
    df, sampling rate: DataFrame, int
        The AcqKnowledge file converted to a dataframe and its sampling rate.

    See Also
    --------
    signal_resample

    Example
    ----------
    >>> import neurokit2 as nk
    >>>
    >>> data, sampling_rate = nk.read_acqknowledge('file.acq')
    """



    # Check filename
    if ".acq" not in filename:
        filename += ".acq"

    if os.path.exists(filename) is False:
        raise ValueError("NeuroKit error: read_acqknowledge(): couldn't find" \
                         " the following file: " + filename)


    # Read file
    file = bioread.read(filename)

    # Get desired frequency
    if sampling_rate == "max":
        freq_list = []
        for channel in file.named_channels:
            freq_list.append(file.named_channels[channel].samples_per_second)
        sampling_rate = np.max(freq_list)

    # Loop through channels
    data = {}
    for channel in file.named_channels:
        signal = np.array(file.named_channels[channel].data)

        # Fill signal interruptions
        if impute_missing is True:
            if np.isnan(np.sum(signal)):
                signal = pd.Series(signal).fillna(method="pad").values

        # Resample if necessary
        if file.named_channels[channel].samples_per_second != sampling_rate:
            signal = signal_resample(signal,
                                     sampling_rate=file.named_channels[channel].samples_per_second,
                                     desired_sampling_rate=sampling_rate,
                                     method=resample_method)
        data[channel] = signal

    # Sanitize lengths
    lengths = []
    for channel in data:
       lengths += [len(data[channel])]
    if len(set(lengths)) > 1:  # If different lengths
        length = pd.Series(lengths).mode()[0]  # Find most common (target length)
        for channel in data:
            if len(data[channel]) > length:
                data[channel] = data[channel][0:length]
            if len(data[channel]) < length:
                  data[channel] = np.concatenate([data[channel],
                                                 np.full((length-len(data[channel])), data[channel][-1])])

    # Final dataframe
    df = pd.DataFrame(data)
    return(df, sampling_rate)


# -*- coding: utf-8 -*-
import numpy as np
import neurokit2 as nk
import matplotlib.pyplot as plt

import biosppy

def test_ecg_simulate():

    ecg1 = nk.ecg_simulate(duration=20, length=5000, method="simple", noise=0)
    assert len(ecg1) == 5000

    ecg2 = nk.ecg_simulate(duration=20, length=5000, heart_rate=500)
#    pd.DataFrame({"ECG1":ecg1, "ECG2": ecg2}).plot()
#    pd.DataFrame({"ECG1":ecg1, "ECG2": ecg2}).hist()
    assert (len(nk.signal_findpeaks(ecg1, height_min=0.6)[0]) <
            len(nk.signal_findpeaks(ecg2, height_min=0.6)[0]))

    ecg3 = nk.ecg_simulate(duration=10, length=5000)
#    pd.DataFrame({"ECG1":ecg1, "ECG3": ecg3}).plot()
    assert (len(nk.signal_findpeaks(ecg2, height_min=0.6)[0]) >
            len(nk.signal_findpeaks(ecg3, height_min=0.6)[0]))



def test_ecg_clean():

    sampling_rate = 1000
    noise = 0.05

    ecg = nk.ecg_simulate(sampling_rate=sampling_rate, noise=noise)
    ecg_cleaned_nk = nk.ecg_clean(ecg, sampling_rate=sampling_rate,
                                  method="neurokit")

    assert ecg.size == ecg_cleaned_nk.size

    # Assert that highpass filter with .5 Hz lowcut was applied.
    fft_raw = np.abs(np.fft.rfft(ecg))
    fft_nk = np.abs(np.fft.rfft(ecg_cleaned_nk))

    freqs = np.fft.rfftfreq(ecg.size, 1 / sampling_rate)

    assert np.sum(fft_raw[freqs < .5]) > np.sum(fft_nk[freqs < .5])

    # Compariosn to biosppy (https://github.com/PIA-Group/BioSPPy/blob/e65da30f6379852ecb98f8e2e0c9b4b5175416c3/biosppy/signals/ecg.py#L69)
    ecg_biosppy = nk.ecg_clean(ecg, sampling_rate=sampling_rate, method="biosppy")
    original, _, _ = biosppy.tools.filter_signal(signal=ecg,
                                                 ftype='FIR',
                                                 band='bandpass',
                                                 order=int(0.3 * sampling_rate),
                                                 frequency=[3, 45],
                                                 sampling_rate=sampling_rate)
    assert np.allclose((ecg_biosppy - original).mean(), 0, atol=1e-6)

def test_ecg_findpeaks():

    sampling_rate = 1000
    noise = 0.05

    ecg = nk.ecg_simulate(sampling_rate=sampling_rate, noise=noise)
    ecg_cleaned_nk = nk.ecg_clean(ecg, sampling_rate=sampling_rate,
                                  method="neurokit")
    signals, info = nk.ecg_findpeaks(ecg_cleaned_nk, method="neurokit")

    assert signals.shape == (10000, 1)
    assert np.allclose(signals["ECG_Peaks"].values.sum(dtype=np.int64), 11)
    assert info["ECG_Peaks"].shape[0] == 11
    assert np.allclose(info["ECG_Peaks"].sum(dtype=np.int64), 56552, atol=1)


def test_ecg_rate():

    sampling_rate = 1000
    noise = 0.05

    ecg = nk.ecg_simulate(sampling_rate=sampling_rate, noise=noise)
    ecg_cleaned_nk = nk.ecg_clean(ecg, sampling_rate=sampling_rate,
                                  method="neurokit")
    signals, info = nk.ecg_findpeaks(ecg_cleaned_nk, method="neurokit")

    # Test with dictionary.
    test_length = 30
    data = nk.ecg_rate(peaks=info, sampling_rate=sampling_rate,
                       desired_length=test_length)

    assert data.shape == (test_length, 1)
    assert np.abs(data["ECG_Rate"].mean() - 70.5) < 0.1

    # Test with DataFrame.
    data = nk.ecg_rate(peaks=signals, sampling_rate=sampling_rate)
    assert data.shape == (ecg.size, 1)
    assert np.abs(data["ECG_Rate"].mean() - 70.5) < 0.2


def test_ecg_process():

    sampling_rate = 1000
    noise = 0.05

    ecg = nk.ecg_simulate(sampling_rate=sampling_rate, noise=noise)
    signals, info = nk.ecg_process(ecg, sampling_rate=sampling_rate,
                                   method="neurokit")
    # Only check array dimensions and column names since functions called by
    # ecg_process have already been unit tested
    assert np.array(["ECG_Raw", "ECG_Clean", "ECG_Peaks", "ECG_Rate"]) in signals.columns.values


def test_ecg_plot():

    ecg = nk.ecg_simulate(duration=60, heart_rate=70, noise=0.05)

    ecg_summary, _ = nk.ecg_process(ecg, sampling_rate=1000, method="neurokit")

    # Plot data over samples.
    nk.ecg_plot(ecg_summary)
    # This will identify the latest figure.
    fig = plt.gcf()
    assert len(fig.axes) == 2
    titles = ["Raw and Cleaned ECG with R-peaks",
              "Heart Rate"]
    for (ax, title) in zip(fig.get_axes(), titles):
        assert ax.get_title() == title
    assert fig.get_axes()[1].get_xlabel() == "Samples"
    plt.close(fig)

    # Plot data over seconds.
    nk.ecg_plot(ecg_summary, sampling_rate=1000)
    # This will identify the latest figure.
    fig = plt.gcf()
    assert len(fig.axes) == 2
    titles = ["Raw and Cleaned ECG with R-peaks",
              "Heart Rate"]
    for (ax, title) in zip(fig.get_axes(), titles):
        assert ax.get_title() == title
    assert fig.get_axes()[1].get_xlabel() == "Time (seconds)"
    plt.close(fig)



import numpy as np
import pandas as pd
import neurokit2 as nk

import os

path_data = os.path.join(os.path.dirname(os.path.dirname(os.path.abspath(__file__))), "data")

# =============================================================================
# Data
# =============================================================================


def test_read_acqknowledge():

    df, sampling_rate = nk.read_acqknowledge(os.path.join(path_data, "example_acqnowledge.acq"), sampling_rate=2000)
    assert sampling_rate == 2000



import numpy as np
import pandas as pd
import neurokit2 as nk
import scipy.signal
import biosppy


# =============================================================================
# Signal
# =============================================================================

def test_signal_smooth():

    # TODO: test kernels other than "boxcar"
    signal = np.cos(np.linspace(start=0, stop=20, num=1000))
    smooth1 = nk.signal_smooth(signal, kernel="boxcar", size=100)
    smooth2 = nk.signal_smooth(signal, kernel="boxcar", size=500)
    # assert that the signal's amplitude is attenuated more with wider kernels
    assert np.allclose(np.std(smooth1), 0.6044, atol=0.00001)
    assert np.allclose(np.std(smooth2), 0.1771, atol=0.0001)


def test_signal_binarize():

    signal = np.cos(np.linspace(start=0, stop=20, num=1000))
    binary = nk.signal_binarize(signal)
    assert len(binary) == 1000

    binary = nk.signal_binarize(list(signal))
    assert len(binary) == 1000


def test_signal_resample():

    signal = np.cos(np.linspace(start=0, stop=20, num=50))

    downsampled_interpolation = nk.signal_resample(signal,
                                                   method="interpolation",
                                                   sampling_rate=1000,
                                                   desired_sampling_rate=500)
    downsampled_numpy = nk.signal_resample(signal, method="numpy",
                                           sampling_rate=1000,
                                           desired_sampling_rate=500)
    downsampled_pandas = nk.signal_resample(signal, method="pandas",
                                            sampling_rate=1000,
                                            desired_sampling_rate=500)
    downsampled_fft = nk.signal_resample(signal, method="FFT",
                                         sampling_rate=1000,
                                         desired_sampling_rate=500)
    downsampled_poly = nk.signal_resample(signal, method="poly",
                                          sampling_rate=1000,
                                          desired_sampling_rate=500)

    # Upsample
    upsampled_interpolation = nk.signal_resample(downsampled_interpolation,
                                                 method="interpolation",
                                                 sampling_rate=500,
                                                 desired_sampling_rate=1000)
    upsampled_numpy = nk.signal_resample(downsampled_numpy, method="numpy",
                                         sampling_rate=500,
                                         desired_sampling_rate=1000)
    upsampled_pandas = nk.signal_resample(downsampled_pandas, method="pandas",
                                          sampling_rate=500,
                                          desired_sampling_rate=1000)
    upsampled_fft = nk.signal_resample(downsampled_fft, method="FFT",
                                       sampling_rate=500,
                                       desired_sampling_rate=1000)
    upsampled_poly = nk.signal_resample(downsampled_poly, method="poly",
                                        sampling_rate=500,
                                        desired_sampling_rate=1000)

    # Check
    rez = pd.DataFrame({"Interpolation": upsampled_interpolation - signal,
                        "Numpy": upsampled_numpy - signal,
                        "Pandas": upsampled_pandas - signal,
                        "FFT": upsampled_fft - signal,
                        "Poly": upsampled_poly - signal})
    assert np.allclose(np.mean(rez.mean()), 0.0001, atol=0.0001)


def test_signal_detrend():

    signal = np.cos(np.linspace(start=0, stop=10, num=1000))  # Low freq
    signal += np.cos(np.linspace(start=0, stop=100, num=1000))  # High freq
    signal += 3  # Add baseline

    rez_nk = nk.signal_detrend(signal, order=1)
    rez_scipy = scipy.signal.detrend(signal, type="linear")
    assert np.allclose(np.mean(rez_nk - rez_scipy), 0, atol=0.000001)

    rez_nk = nk.signal_detrend(signal, order=0)
    rez_scipy = scipy.signal.detrend(signal, type="constant")
    assert np.allclose(np.mean(rez_nk - rez_scipy), 0, atol=0.000001)


def test_signal_filter():

    signal = np.cos(np.linspace(start=0, stop=10, num=1000)) # Low freq
    signal += np.cos(np.linspace(start=0, stop=100, num=1000)) # High freq
    filtered = nk.signal_filter(signal, highcut=10)
    assert np.std(signal) > np.std(filtered)

def test_signal_interpolate():

    x_axis = np.linspace(start=10, stop=30, num=10)
    signal = np.cos(x_axis)

    interpolated = nk.signal_interpolate(signal, desired_length=1000)
    assert len(interpolated) == 1000

    new_x = np.linspace(start=0, stop=40, num=1000)
    interpolated = nk.signal_interpolate(signal,
                                         desired_length=1000,
                                         x_axis=x_axis,
                                         new_x=new_x)
    assert len(interpolated) == 1000
    assert interpolated[0] == signal[0]


def test_signal_findpeaks():

    signal1 = np.cos(np.linspace(start=0, stop=30, num=1000))
    peaks1, info1 = nk.signal_findpeaks(signal1)

    signal2 = np.concatenate([np.arange(0, 20, 0.1), np.arange(17, 30, 0.1),
                              np.arange(30, 10, -0.1)])
    peaks2, info2 = nk.signal_findpeaks(signal2)
    assert len(peaks1) > len(peaks2)


def test_signal_merge():

    signal1 = np.cos(np.linspace(start=0, stop=10, num=100))
    signal2 = np.cos(np.linspace(start=0, stop=20, num=100))

    signal = nk.signal_merge(signal1, signal2, time1=[0, 10], time2=[-5, 5])
    assert len(signal) == 150
    assert signal[0] == signal2[0] + signal2[0]



# -*- coding: utf-8 -*-
import numpy as np
import pandas as pd

from ..events.events_plot import events_plot


def signal_plot(signal):
    """Plot signal with events as vertical lines.

    Parameters
    ----------
    signal : array or DataFrame
        Signal array (can be a dataframe with many signals).


    Examples
    ----------
    >>> import numpy as np
    >>> import pandas as pd
    >>> import neurokit2 as nk
    >>>
    >>> signal = np.cos(np.linspace(start=0, stop=20, num=1000))
    >>> nk.signal_plot(signal)
    >>>
    >>> data = pd.DataFrame({"Signal1": np.cos(np.linspace(start=0, stop=20, num=1000)),
                             "Signal2": np.sin(np.linspace(start=0, stop=20, num=1000)),
                             "Signal3": nk.signal_binarize(np.cos(np.linspace(start=0, stop=40, num=1000)))})
    >>> nk.signal_plot(data)
    """
    # Sanitize format
    if isinstance(signal, pd.DataFrame) is False:
        if len(np.array(signal).shape) > 1:
            signal = pd.DataFrame(np.array(signal).T)
            signal.columns = np.char.add(np.full(len(signal.columns), "Signal"), np.array(np.arange(len(signal.columns)) + 1, dtype=np.str))
        else:
            signal = pd.DataFrame({"Signal": signal})

    # Guess continuous and events columns
    continuous_columns = list(signal.columns.values)
    events_columns = []
    for col in signal.columns:
        vector = signal[col]
        if vector.nunique() == 2:
            indices = np.where(vector == np.max(vector.unique()))
            if bool(np.any(np.diff(indices) == 1)) is False:
                events_columns.append(col)
                continuous_columns.remove(col)

    # Plot accordingly
    if len(events_columns) > 0:
        events = []
        for col in events_columns:
            events.append([signal[col].values])

        events_plot(events, signal=signal[continuous_columns])
    else:
        signal[continuous_columns].plot()



# -*- coding: utf-8 -*-
import pandas as pd
import numpy as np

from ..signal import signal_resample


def emg_simulate(duration=10, length=None, sampling_rate=1000, noise=0.01,
                 n_bursts=1, duration_bursts=1.0, random_state=42):
    """Simulate an EMG signal

    Generate an artificial (synthetic) EMG signal of a given duration and sampling rate.

    Parameters
    ----------
    duration : int
        Desired recording length in seconds.
    sampling_rate, length : int
        The desired sampling rate (in Hz, i.e., samples/second) or the desired
        length of the signal (in samples).
    noise : float
        Noise level (gaussian noise).
    n_bursts : int
        Desired number of bursts of activity (active muscle periods).
    duration_bursts : float or list
        Duration of the bursts. Can be a float (each burst will have the same
        duration) or a list of durations for each bursts.
    random_state : int
        Seed for the random number generator.

    Returns
    ----------
    array
        Vector containing the EMG signal.

    Examples
    ----------
    >>> import neurokit as nk
    >>> import pandas as pd
    >>>
    >>> emg = nk.emg_simulate(duration=10, n_bursts=3)
    >>> nk.signal_plot(emg)

    See Also
    --------
    ecg_simulate, rsp_simulate, eda_simulate, ppg_simulate


    References
    -----------
    This function is based on `this script <https://scientificallysound.org/2016/08/11/python-analysing-emg-signals-part-1/>`_.
    """
    # Seed the random generator for reproducible results
    np.random.seed(random_state)

    # Generate number of samples automatically if length is unspecified
    if length is None:
        length = duration * sampling_rate

    # Sanity checks
    if isinstance(duration_bursts, int) or isinstance(duration_bursts, float):
        duration_bursts = np.repeat(duration_bursts, n_bursts)

    if len(duration_bursts) > n_bursts:
        raise ValueError("NeuroKit error: emg_simulate(): 'duration_bursts'" \
                         " cannot be longer than the value of 'n_bursts'")

    total_duration_bursts = np.sum(duration_bursts)
    if total_duration_bursts > duration:
        raise ValueError("NeuroKit error: emg_simulate(): The total duration" \
                         " of bursts cannot exceed the total duration")

    # Generate bursts
    bursts = []
    for burst in range(n_bursts):
        bursts += [list(np.random.uniform(-1, 1, size=int(1000*duration_bursts[burst])) + 0.08)]

    # Generate quiet
    n_quiet = n_bursts + 1  # number of quiet periods (in between bursts)
    duration_quiet = (duration - total_duration_bursts) / n_quiet  # duration of each quiet period
    quiets = []
    for quiet in range(n_quiet):
        quiets += [list(np.random.uniform(-0.05, 0.05, size=int(1000*duration_quiet)) + 0.08)]

    # Merge the two
    emg = []
    for i in range(len(quiets)):
        emg += quiets[i]
        if i < len(bursts):
            emg += bursts[i]
    emg = np.array(emg)

    # Add random (gaussian distributed) noise
    emg += np.random.normal(0, noise, len(emg))

    # Resample
    emg = signal_resample(emg,
                          sampling_rate=1000,
                          desired_length=length,
                          desired_sampling_rate=sampling_rate)

    return(emg)



# -*- coding: utf-8 -*-
import numpy as np






def events_to_mne(events, conditions=None):
    """Create `MNE <https://mne.tools/stable/index.html>`_ compatible events for integration with M/EEG..

    Parameters
    ----------
    events : list, ndarray or dict
        Events onset location. If a dict is passed (e.g., from 'events_find()'), will select only the 'Onset' list.
    conditions : list
        A list of same length as events containing the stimuli types/conditions.


    Returns
    ----------
    (events, event_id) : tuple
        MNE-formatted events, that can be added via 'raw.add_events(events), and a dictionary with event's names.

    See Also
    --------
    events_find

    Examples
    ----------
    >>> import numpy as np
    >>> import pandas as pd
    >>> import neurokit2 as nk
    >>>
    >>> signal = np.cos(np.linspace(start=0, stop=20, num=1000))
    >>> events = nk.events_find(signal)
    >>> events, event_id = nk.events_to_mne(events)
    >>> event_id
    {'Event': 0}
    """

    if isinstance(events, dict):
        events = events["Onset"]

    event_id = {}

    if conditions is None:
        conditions = ["Event"] * len(events)

    # Sanity check
    if len(conditions) != len(events):
        raise ValueError("NeuroKit error: events_to_mne(): 'conditions' argument of different length than event onsets.")


    event_names = list(set(conditions))
    event_index = list(range(len(event_names)))
    for i in enumerate(event_names):
        conditions = [event_index[i[0]] if x==i[1] else x for x in conditions]
        event_id[i[1]] = event_index[i[0]]

    events = np.array([events, [0]*len(events), conditions]).T
    return(events, event_id)



import numpy as np
import pandas as pd
import neurokit2 as nk

# =============================================================================
# Events
# =============================================================================



def test_events_find():

    signal = np.cos(np.linspace(start=0, stop=20, num=1000))
    events = nk.events_find(signal)
    assert list(events["Onset"]) == [0, 236, 550, 864]

    events = nk.events_find(signal, duration_min = 150)
    assert list(events["Onset"]) == [236, 550]

    events = nk.events_find(signal, inter_min = 300)
    assert list(events["Onset"]) == [0, 550, 864]



def test_events_to_mne():

    signal = np.cos(np.linspace(start=0, stop=20, num=1000))
    events = nk.events_find(signal)
    events, event_id = nk.events_to_mne(events)
    assert event_id == {'Event': 0}



def test_events_plot():

    signal = np.cos(np.linspace(start=0, stop=20, num=1000))
    events = nk.events_find(signal)
    data = nk.events_plot(events, signal, show=False)
    assert len(data['Event_Onset']) == 1000


# -*- coding: utf-8 -*-
import numpy as np
import pandas as pd
from ..signal import signal_interpolate


def rsp_amplitude(rsp_signal, extrema, desired_length=None):
    """Compute respiratory amplitude.

    Compute respiratory amplitude given the raw respiration signal and its
    extrema.

    Parameters
    ----------
    rsp_signal : list, array or Series
        The raw respiration channel (as measured, for instance, by a
        respiration belt).
    extrema : DataFrame, dict
        The extrema (i.e., inhalation peaks and exhalation troughs) found in
        the raw respiration channel. Must be one of the data containers
        returned by `rsp_findpeaks()`.
    desired_length : int
        By default, the returned respiratory amplitude has the same number of
        elements as inhalation peaks in `extrema`. If set to an integer, the
        returned amplitude will be interpolated between inhalation peaks over
        `desired_length` samples. Has not effect if a DataFrame is passed in as
        the `extrema` argument. In the latter case the respiratory amplitude
        will be interpolated over the entire duration of `rsp_signal`.

    Returns
    -------
    signals : DataFrame
        A DataFrame containing the respiratory amplitude accessible with the
        key "RSP_Amplitude"

    See Also
    --------
    rsp_clean, rsp_findpeaks, rsp_rate, rsp_process, rsp_plot

    Examples
    --------
    >>> import neurokit2 as nk
    >>>
    >>> rsp = nk.rsp_simulate(duration=90, respiratory_rate=15)
    >>> cleaned = nk.rsp_clean(rsp, sampling_rate=1000)
    >>> extrema, info = nk.rsp_findpeaks(cleaned)
    >>>
    >>> amplitude = nk.rsp_amplitude(rsp, extrema)
    """
    # Sanity checks
    error_msg0 = ("NeuroKit error: Please provide one of the containers "
                  "returned by `rsp_findpeaks()` as `extrema` argument.")
    error_msg1 = ("NeuroKit error: Please provide one of the containers "
                  "returned by `rsp_findpeaks()` as `extrema` argument and do "
                  "not modify its content.")
    if isinstance(extrema, dict):
        try:
            peaks = extrema["RSP_Peaks"]
            troughs = extrema["RSP_Troughs"]
        except (TypeError, KeyError):
            raise TypeError(error_msg0)
    elif isinstance(extrema, pd.DataFrame):
        try:
            desired_length = extrema["RSP_Peaks"].size
            peaks = np.where(extrema["RSP_Peaks"] == 1)[0]
            troughs = np.where(extrema["RSP_Troughs"] == 1)[0]
        except (TypeError, KeyError):
            raise TypeError(error_msg0)
    else:
        raise TypeError(error_msg0)

    # To consistenty calculate amplitude, peaks and troughs must have the same
    # number of elements, and the first trough must precede the first peak.
    if peaks.size != troughs.size:
        raise TypeError(error_msg1)
    if peaks[0] <= troughs[0]:
        raise TypeError(error_msg1)


    # Determine length of final signal to return.
    if desired_length is None:
        desired_length = peaks.size

    # Calculate amplitude in units of the raw signal, based on vertical
    # difference of each peak to the preceding trough.
    amplitude = np.abs(rsp_signal[peaks] - rsp_signal[troughs])

    # Interpolate amplitude to desired_length samples.
    amplitude = signal_interpolate(amplitude, x_axis=peaks,
                                   desired_length=desired_length)

    signals = pd.DataFrame(amplitude, columns=["RSP_Amplitude"])

    return signals



import numpy as np
import pandas as pd
import neurokit2 as nk

# =============================================================================
# EDA
# =============================================================================

def test_eda_simulate():

    eda1 = nk.eda_simulate(duration=10, length=None, n_scr=1)
    assert len(nk.signal_findpeaks(eda1, height_min=0.6)[0]) == 1

    eda2 = nk.eda_simulate(duration=10, length=None, n_scr=5)
    assert len(nk.signal_findpeaks(eda2, height_min=0.6)[0]) == 5
#   pd.DataFrame({"EDA1": eda1, "EDA2": eda2}).plot()

    assert len(nk.signal_findpeaks(eda2, height_min=0.6)[0]) > len(nk.signal_findpeaks(eda1, height_min=0.6)[0])



# -*- coding: utf-8 -*-
import numpy as np
import pandas as pd

from .signal_resample import signal_resample
from ..misc import listify


def signal_distord(signal, sampling_rate=1000, noise_amplitude=0.1, noise_frequency=100, noise_shape="laplace", powerline_amplitude=0, powerline_frequency=50):
    """Signal distortion.

    Add noise of a given frequency, amplitude and shape to a signal.

    Parameters
    ----------
    signal : list, array or Series
        The signal channel in the form of a vector of values.
    sampling_rate : int
        The sampling frequency of the signal (in Hz, i.e., samples/second).
    noise_amplitude : float
        The amplitude of the noise (the scale of the random function, relative
        to the standard deviation of the signal).
    noise_frequency : int
        The frequency of the noise (in Hz, i.e., samples/second).
    noise_shape : str
        The shape of the noise. Can be one of 'laplace' (default) or 'gaussian'.

    Returns
    -------
    array
        Vector containing the distorted signal.

    Examples
    --------
    >>> import numpy as np
    >>> import pandas as pd
    >>> import neurokit2 as nk
    >>>
    >>> signal = np.cos(np.linspace(start=0, stop=10, num=10000))
    >>> signals = pd.DataFrame({
            "Freq100": nk.signal_distord(signal, noise_frequency=200),
            "Freq50": nk.signal_distord(signal, noise_frequency=50),
            "Freq10": nk.signal_distord(signal, noise_frequency=10),
            "Freq5": nk.signal_distord(signal, noise_frequency=5),
            "Raw": signal})
    >>> signals.plot()
    >>>
    >>> distorted = nk.signal_distord(signal, noise_amplitude=[0.3, 0.2, 0.1], noise_frequency=[5, 10, 20], powerline_amplitude=0.05)
    >>> nk.signal_plot(distorted)
    """
    # Basic noise
    noise = _signal_distord_noise_multifrequency(signal,
                                                 signal_sd=np.std(signal, ddof=1),
                                                 sampling_rate=sampling_rate,
                                                 noise_amplitude=noise_amplitude,
                                                 noise_frequency=noise_frequency,
                                                 noise_shape=noise_shape)

    # Powerline noise
    if powerline_amplitude > 0:
        noise += _signal_distord_powerline(signal,
                                           signal_sd=np.std(signal, ddof=1),
                                           sampling_rate=sampling_rate,
                                           powerline_amplitude=powerline_amplitude,
                                           powerline_frequency=powerline_frequency)
    distorted = signal + noise

    return distorted







# =============================================================================
# Internals
# =============================================================================

def _signal_distord_powerline(signal, signal_sd=None, sampling_rate=1000, powerline_frequency=50, powerline_amplitude=0.1):
    freqs = list(np.arange(powerline_frequency, sampling_rate, powerline_frequency))
    noise = _signal_distord_noise_multifrequency(signal,
                                                 signal_sd=signal_sd,
                                                 sampling_rate=sampling_rate,
                                                 noise_amplitude=powerline_amplitude,
                                                 noise_frequency=freqs,
                                                 noise_shape="gaussian")
    return noise




def _signal_distord_noise_multifrequency(signal, signal_sd=None, sampling_rate=1000, noise_amplitude=0.1, noise_frequency=100, noise_shape="laplace"):
    duration = len(signal) / sampling_rate

    noise = np.zeros(len(signal))
    params = listify(noise_amplitude=noise_amplitude, noise_frequency=noise_frequency, noise_shape=noise_shape)
    for i in range(len(params["noise_amplitude"])):
        if params["noise_frequency"][i] <= sampling_rate:  # Skip noise of higher freq than recording

            # Parameters
            noise_duration = int(duration * params["noise_frequency"][i])
            if signal_sd is None:
                amplitude = params["noise_amplitude"][i]
            else:
                amplitude = params["noise_amplitude"][i] * signal_sd
            shape = params["noise_shape"][i]

            # Generate noise
            noise = _signal_distord_noise(signal, noise_duration, amplitude, shape)
            noise += noise
    return noise



def _signal_distord_noise(signal, noise_duration, noise_amplitude=0.1, noise_shape="laplace"):

    if noise_shape in ["normal", "gaussian"]:
        noise = np.random.normal(0, noise_amplitude, noise_duration)
    elif noise_shape == "laplace":
        noise = np.random.laplace(0, noise_amplitude, noise_duration)
    else:
        raise ValueError("NeuroKit error: signal_distord(): 'noise_shape' "
                         "should be one of 'gaussian' or 'laplace'.")

    noise = signal_resample(noise, desired_length=len(signal), method="interpolation")
    return noise



# -*- coding: utf-8 -*-
import numpy as np

from ..signal import signal_merge
from ..signal import signal_distord

def eda_simulate(duration=10, length=None, sampling_rate=1000, noise=0.01,
                 n_scr=1, drift=-0.01, random_state=42):
    """Simulate Electrodermal Activity (EDA) signal.

    Generate an artificial (synthetic) EDA signal of a given duration and sampling rate.

    Parameters
    ----------
    duration : int
        Desired recording length in seconds.
    sampling_rate, length : int
        The desired sampling rate (in Hz, i.e., samples/second) or the desired
        length of the signal (in samples).
    noise : float
        Noise level (amplitude of the laplace noise).
    n_scr : int
        Desired number of skin conductance responses (SCRs), i.e., peaks.
    drift : float or list
        The slope of a linear drift of the signal.
    random_state : int
        Seed for the random number generator.

    Returns
    ----------
    array
        Vector containing the EDA signal.

    Examples
    ----------
    >>> import neurokit as nk
    >>> import pandas as pd
    >>>
    >>> eda = nk.eda_simulate(duration=10, n_scr=3)
    >>> nk.signal_plot(eda)

    See Also
    --------
    ecg_simulate, rsp_simulate, emg_simulate, ppg_simulate


    References
    -----------
    - Bach, D. R., Flandin, G., Friston, K. J., & Dolan, R. J. (2010). Modelling event-related skin conductance responses. International Journal of Psychophysiology, 75(3), 349-356.
    """
    # Seed the random generator for reproducible results
    np.random.seed(random_state)

    # Generate number of samples automatically if length is unspecified
    if length is None:
        length = duration * sampling_rate

    eda = np.full(length, 1.0)
    eda += (drift * np.linspace(0, duration, length))
    time = [0, duration]

    start_peaks = np.linspace(0, duration, n_scr, endpoint=False)

    for start_peak in start_peaks:
        relative_time_peak = np.abs(np.random.normal(0, 5, size=1)) + 3.0745
        scr = _eda_simulate_scr(sampling_rate=sampling_rate,
                                time_peak=relative_time_peak)
        time_scr = [start_peak, start_peak+9]
        if time_scr[0] < 0:
            scr = scr[int(np.round(np.abs(time_scr[0])*sampling_rate))::]
            time_scr[0] = 0
        if time_scr[1] > duration:
            scr = scr[0:int(np.round((duration - time_scr[0])*sampling_rate))]
            time_scr[1] = duration

        eda = signal_merge(signal1=eda, signal2=scr, time1=time, time2=time_scr)

    # Add random noise
    if noise > 0:
        eda = signal_distord(eda,
                             sampling_rate=sampling_rate,
                             noise_amplitude=noise,
                             noise_frequency=[5, 10, 100],
                             noise_shape="laplace")
    return eda






def _eda_simulate_scr(sampling_rate=1000,
                      length=None,
                      time_peak=3.0745,
                      rise=0.7013,
                      decay=[3.1487, 14.1257]):
    """Simulate a canonical skin conductance response (SCR)

    Based on `Bach (2010) <https://sourceforge.net/p/scralyze/code/HEAD/tree/branches/version_b2.1.8/scr_bf_crf.m#l24>`_

    Parameters
    -------------
    time_peak : float
        Time to peak.
    rise : float
        Variance of rise defining gaussian.
    decay : list
        Decay constants.

    Examples
    --------
    >>> scr1 = _eda_simulate_canonical(time_peak=3.0745)
    >>> scr2 = _eda_simulate_canonical(time_peak=10)
    >>> pd.DataFrame({"SCR1": scr1, "SCR2": scr2}).plot()
    """
    if length is None:
        length = 9*sampling_rate
    t = np.linspace(sampling_rate/10000, 90, length)

    gt = np.exp(-((t - time_peak)**2)/(2*rise**2))
    ht = np.exp(-t/decay[0]) + np.exp(-t/decay[1])

    ft = np.convolve(gt, ht)
    ft = ft[0:len(t)]
    ft = ft/np.max(ft)
    return ft



"""Submodule for NeuroKit."""

from .signal_binarize import signal_binarize
from .signal_resample import signal_resample
from .signal_zerocrossings import signal_zerocrossings
from .signal_smooth import signal_smooth
from .signal_filter import signal_filter
from .signal_psd import signal_psd
from .signal_distord import signal_distord
from .signal_interpolate import signal_interpolate
from .signal_detrend import signal_detrend
from .signal_findpeaks import signal_findpeaks
from .signal_merge import signal_merge
from .signal_plot import signal_plot



import numpy as np
import pandas as pd
import neurokit2 as nk

import mne


# =============================================================================
# EEG
# =============================================================================


def test_eeg_add_channel():


    raw = mne.io.read_raw_fif(mne.datasets.sample.data_path() + '/MEG/sample/sample_audvis_raw.fif', preload=True)

    # len(channel) > len(raw)
    ecg1 = nk.ecg_simulate(length=170000)

    # sync_index_raw > sync_index_channel
    raw1 = nk.eeg_add_channel(raw.copy(), ecg1, channel_type="ecg", sync_index_raw=100, sync_index_channel=0)
    df1 = raw1.to_data_frame()

    # test if the column of channel is added
    assert len(df1.columns) == 377

    # test if the NaN is appended properly to the added channel to account for difference in distance between two signals
    sync_index_raw=100
    sync_index_channel=0
    for i in df1["Added_Channel"].head(abs(sync_index_channel - sync_index_raw)):
        assert np.isnan(i)
    assert np.isfinite(df1["Added_Channel"].iloc[abs(sync_index_channel - sync_index_raw)])





    # len(channel) < len(raw)
    ecg2 = nk.ecg_simulate(length=166790)

    # sync_index_raw < sync_index_channel
    raw2 = nk.eeg_add_channel(raw.copy(), ecg2, channel_type="ecg", sync_index_raw=0, sync_index_channel=100)
    df2 = raw2.to_data_frame()

    # test if the column of channel is added
    assert len(df2.columns) == 377

    # test if the NaN is appended properly to the added channel to account for difference in distance between two signals + difference in length
    sync_index_raw=0
    sync_index_channel=100
    for i in df2["Added_Channel"].tail(abs(sync_index_channel - sync_index_raw) + (len(raw) - len(ecg2))):
        assert np.isnan(i)
    assert np.isfinite(df2["Added_Channel"].iloc[-abs(sync_index_channel - sync_index_raw) - (len(raw) - len(ecg2)) - 1])


# -*- coding: utf-8 -*-
import numpy as np
import pandas as pd
import matplotlib.pyplot as plt


def ecg_plot(ecg_signals, sampling_rate=None):
    """Visualize ECG data.

    Parameters
    ----------
    ecg_signals : DataFrame
        DataFrame obtained from `ecg_process()`.
    sampling_rate : int
        The sampling frequency of the ECG (in Hz, i.e., samples/second). Needs
        to be supplied if the data should be plotted over time in seconds.
        Otherwise the data is plotted over samples. Defaults to None.

    Examples
    --------
    >>> import neurokit2 as nk
    >>>
    >>> ecg = nk.ecg_simulate(duration=15, sampling_rate=1000, heart_rate=80)
    >>> signals, info = nk.ecg_process(ecg, sampling_rate=1000)
    >>> nk.ecg_plot(signals)

    See Also
    --------
    ecg_process
    """

    # Sanity-check input.
    if not isinstance(ecg_signals, pd.DataFrame):
        print("NeuroKit error: The `ecg_signals` argument must be the "
              "DataFrame returned by `ecg_process()`.")
        return

    # Determine what to display on the x-axis.
    if sampling_rate is not None:
        x_axis = np.linspace(0, ecg_signals.shape[0] / sampling_rate,
                             ecg_signals.shape[0])
    else:
        x_axis = np.arange(0, ecg_signals.shape[0])

    # Extract R-peaks.
    peaks = np.where(ecg_signals["ECG_Peaks"] == 1)[0]

    # Prepare figure.
    fig, (ax0, ax1) = plt.subplots(nrows=2, ncols=1, sharex=True)
    if sampling_rate is not None:
        ax1.set_xlabel("Time (seconds)")
    elif sampling_rate is None:
        ax1.set_xlabel("Samples")

    fig.suptitle("Electrocardiogram (ECG) Processing", fontweight="bold")
    plt.subplots_adjust(hspace=0.2)

    # Plot cleaned and raw ECG as well as R-peaks.
    ax0.set_title("Raw and Cleaned ECG with R-peaks")

    ax0.plot(x_axis, ecg_signals["ECG_Raw"], color='#B0BEC5', label='Raw',
             zorder=1)
    ax0.plot(x_axis, ecg_signals["ECG_Clean"], color='#E91E63',
             label="Cleaned", zorder=1)
    ax0.scatter(x_axis[peaks], ecg_signals["ECG_Clean"][peaks], color="#FFC107",
                label="R-peaks", zorder=2)

    ax0.legend(loc="upper right")

    # Plot heart rate.
    ax1.set_title("Heart Rate")
    ax1.set_ylabel("Beats per minute (bpm)")

    ax1.plot(x_axis, ecg_signals["ECG_Rate"], color="#FF5722", label="Signal")
    rate_mean = ecg_signals["ECG_Rate"].mean()
    ax1.axhline(y=rate_mean, label="Mean", linestyle="--", color="#FF9800")

    ax1.legend(loc="upper right")

    plt.show()
    return fig



# -*- coding: utf-8 -*-
import pandas as pd
import numpy as np

from .utils_phi import _phi
from .utils_get_r import _get_r



def entropy_approximate(signal, order=2, r="default"):
    """Compute the approximate entropy (ApEn).

    Approximate entropy is a technique used to quantify the amount of regularity and the unpredictability of fluctuations over time-series data. The advantages of ApEn include lower computational demand (ApEn can be designed to work for small data samples (< 50 data points) and can be applied in real tim) and less sensitive to noise. However, ApEn is heavily dependent on the record length and lacks relative consistency.


    Parameters
    ----------
    signal : list, array or Series
        The signal channel in the form of a vector of values.
    order : int
        The Embedding dimension (often denoted as 'm'), i.e., the length of compared run of data. Typically 1, 2 or 3.
    r : float
        Tolerance (i.e., filtering level - max absolute difference between segments). If 'default', will be set to 0.2 times the standard deviation of the signal.

    See Also
    --------
    entropy_shannon, entropy_sample, entropy_fuzzy

    Returns
    ----------
    float
        The approximate entropy as float value.


    Examples
    ----------
    >>> import neurokit2 as nk
    >>>
    >>> signal = np.cos(np.linspace(start=0, stop=30, num=100))
    >>> nk.entropy_approximate(signal[0:100])
    0.17364897858477146


    References
    -----------
    - `EntroPy` <https://github.com/raphaelvallat/entropy>`_
    - Sabeti, M., Katebi, S., & Boostani, R. (2009). Entropy and complexity measures for EEG signal classification of schizophrenic and control participants. Artificial intelligence in medicine, 47(3), 263-274.
    """
    r = _get_r(signal, r=r)

    # Get phi
    phi = _phi(signal, order=order, r=r, metric='chebyshev', approximate=True)
    return(np.abs(np.subtract(phi[0], phi[1])))



# -*- coding: utf-8 -*-
import numpy as np
import matplotlib.pyplot as plt


def rsp_plot(rsp_signals, sampling_rate=None):
    """Visualize respiration (RSP) data.

    Parameters
    ----------
    rsp_signals : DataFrame
        DataFrame obtained from `rsp_process()`.

    Examples
    --------
    >>> import neurokit2 as nk
    >>>
    >>> rsp = nk.rsp_simulate(duration=90, respiratory_rate=15)
    >>> signals, info = nk.rsp_process(rsp, sampling_rate=1000)
    >>> nk.rsp_plot(signals)

    See Also
    --------
    rsp_process
    """
    peaks = np.where(rsp_signals["RSP_Peaks"] == 1)[0]
    troughs = np.where(rsp_signals["RSP_Troughs"] == 1)[0]

    if "RSP_Amplitude" in list(rsp_signals.columns):
        fig, (ax0, ax1, ax2) = plt.subplots(nrows=3, ncols=1, sharex=True)
    else:
        fig, (ax0, ax1) = plt.subplots(nrows=2, ncols=1, sharex=True)

    # Determine unit of x-axis.
    last_ax = fig.get_axes()[-1]
    if sampling_rate is not None:
        last_ax.set_xlabel("Seconds")
        x_axis = np.linspace(0, len(rsp_signals) / sampling_rate,
                             len(rsp_signals))
    else:
        last_ax.set_xlabel("Samples")
        x_axis = np.arange(0, len(rsp_signals))

    plt.subplots_adjust(hspace=0.2)

    # Plot cleaned and raw respiration as well as peaks and troughs.
    ax0.set_title("Raw and Cleaned RSP")
    fig.suptitle('Respiration (RSP) Processing', fontweight='bold')

    ax0.plot(x_axis, rsp_signals["RSP_Raw"], color='#B0BEC5', label='Raw',
             zorder=1)
    ax0.plot(x_axis, rsp_signals["RSP_Clean"], color='#2196F3',
             label='Cleaned', zorder=1)

    ax0.scatter(x_axis[peaks], rsp_signals["RSP_Clean"][peaks], color='red',
                label="Inhalation Peaks", zorder=2)
    ax0.scatter(x_axis[troughs], rsp_signals["RSP_Clean"][troughs],
                color='orange', label="Exhalation Troughs", zorder=2)

    ax0.legend(loc='upper right')


    # Plot rate and optionally amplitude.
    ax1.set_title("Breathing Rate")
    ax1.plot(x_axis, rsp_signals["RSP_Rate"], color='#4CAF50', label='Rate')
    rate_mean = np.mean(rsp_signals["RSP_Rate"])
    ax1.axhline(y=rate_mean, label='Mean', linestyle='--', color='#4CAF50')
    ax1.legend(loc='upper right')

    if "RSP_Amplitude" in list(rsp_signals.columns):
        ax2.set_title("Breathing Amplitude")

        ax2.plot(x_axis, rsp_signals["RSP_Amplitude"], color='#009688',
                 label='Amplitude')
        amplitude_mean = np.mean(rsp_signals["RSP_Amplitude"])
        ax2.axhline(y=amplitude_mean, label='Mean', linestyle='--',
                    color='#009688')
        ax2.legend(loc='upper right')

    plt.show()
    return fig



# -*- coding: utf-8 -*-
import pandas as pd
import numpy as np

from .utils_embed import _embed
from .utils_get_r import _get_r


def _entropy_sample(signal, order=2, r="default", n=1, fuzzy=False):
    """
    Internal function adapted from https://github.com/ixjlyons/entro-py/blob/master/entropy.py
    With fixes (https://github.com/ixjlyons/entro-py/pull/2/files) by @CSchoel
    """
    # Sanity checks
    signal = np.array(signal).astype(float)
    N = len(signal)

    phi = [0, 0]  # phi(m), phi(m+1)
    for j in [0, 1]:
        m = order + j
        npat = N - order  # https://github.com/ixjlyons/entro-py/pull/2
#        patterns = np.transpose(_embed(signal, m))
        patterns = np.transpose(_embed(signal, m))[:, :npat]

        if fuzzy:
            patterns -= np.mean(patterns, axis=0, keepdims=True)

#        count = np.zeros(N-m)  # https://github.com/ixjlyons/entro-py/pull/2
#        for i in range(N-m):  # https://github.com/ixjlyons/entro-py/pull/2
        count = np.zeros(npat)
        for i in range(npat):
            if m == 1:
                sub = patterns[i]
            else:
                sub = patterns[:, [i]]
            dist = np.max(np.abs(patterns - sub), axis=0)

            if fuzzy:
                sim = np.exp(-np.power(dist, n) / r)
            else:
                sim = dist < r  # https://github.com/ixjlyons/entro-py/pull/2

            count[i] = np.sum(sim) - 1

#        phi[j] = np.mean(count) / (N-m-1)
        phi[j] = np.mean(count) / (N-order-1)  # https://github.com/ixjlyons/entro-py/pull/2

    return phi



def entropy_fuzzy(signal, order=2, r="default", n=1):
    """
    Calculate the fuzzy entropy (FuzzyEn) of a signal. Adapted from `entro-py <https://github.com/ixjlyons/entro-py/blob/master/entropy.py>`_.

    Parameters
    ----------
    signal : list, array or Series
        The signal channel in the form of a vector of values.
    order : int
        The Embedding dimension (often denoted as 'm'), i.e., the length of compared run of data. Typically 1, 2 or 3.
    r : float
        Tolerance (i.e., filtering level - max absolute difference between segments). If 'default', will be set to 0.2 times the standard deviation of the signal.
    n : float, optional
        Step width of fuzzy exponential function. Larger `n` makes the function
        more rectangular. Usually in the range 1-5 (default is 1).

    Returns
    ----------
    float
        The fuzzy entropy as float value.

    See Also
    --------
    entropy_shannon, entropy_approximate, entropy_sample

    Examples
    ----------
    >>> import neurokit2 as nk
    >>>
    >>> signal = np.cos(np.linspace(start=0, stop=30, num=100))
    >>> nk.entropy_fuzzy(signal[0:100])
    0.5240030033696856
    """
    r = _get_r(signal, r=r)
    phi = _entropy_sample(signal, order=order, r=r, n=1, fuzzy=True)
    return -np.log(np.divide(phi[1], phi[0]))



# -*- coding: utf-8 -*-
import pandas as pd
import numpy as np


from .utils_phi import _phi
from .utils_get_r import _get_r
from .entropy_fuzzy import _entropy_sample






def entropy_sample(signal, order=2, r="default"):
    """
    Calculate the sample entropy (SampEn) of a signal.

    Parameters
    ----------
    signal : list, array or Series
        The signal channel in the form of a vector of values.
    order : int
        The Embedding dimension (often denoted as 'm'), i.e., the length of compared run of data. Typically 1, 2 or 3.
    r : float
        Tolerance (i.e., filtering level - max absolute difference between segments). If 'default', will be set to 0.2 times the standard deviation of the signal.

    See Also
    --------
    entropy_shannon, entropy_approximate, entropy_fuzzy

    Returns
    ----------
    float
        The sample entropy as float value.

    Examples
    ----------
    >>> import neurokit2 as nk
    >>>
    >>> signal = np.cos(np.linspace(start=0, stop=30, num=100))
    >>> nk.entropy_sample(signal[0:100])
    0.281711905465277
    """
    r = _get_r(signal, r=r)

    # entro-py implementation (https://github.com/ixjlyons/entro-py/blob/master/entropy.py):
#    phi = _entropy_sample(signal, order=order, r=r, n=1, fuzzy=False)

    # nolds and Entropy implementation:
    phi = _phi(signal, order=order, r=r, metric='chebyshev', approximate=False)

    return -np.log(np.divide(phi[1], phi[0]))
